# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from zef.core import VT

value_types = [x for x in dir(VT) if isinstance(getattr(VT, x), VT.ValueType)]
value_types.sort()

import os, sys
pyi_filename = "__init__.pyi"

if len(sys.argv) == 2:
    pyi_filename = sys.argv[1]
else:
    assert len(sys.argv) == 1, "Script should be called with no arguments or one argument, the location in which to save the file."

with open(pyi_filename, "w") as file:
    file.write("""# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .value_type import *

""")

    for vtype in value_types:
        file.write(f"{vtype}: ValueType_ = ...\n")



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .. import *
from .._ops import *


def create_eff_to_process_graph_wish(eff: Dict) -> 'FX':
    from .. import internals
    g_process = internals.get_local_process_graph()
    pushable_stream = {'type': FX.Stream.CreatePushableStream} | run

    fields_d = {
        "port" : eff.get('port', 5000),
        "bind_address" : eff.get('bind_address', "0.0.0.0"),                      
        "type" : Val(eff['type']),   
        # TODO fix pipeinto being unserializeable if python objects are found
        # "pipe_into" : Val(eff.get('pipe_into', None)),       
        "PushableStream": pushable_stream.stream_ezefref,
        # "url" : 'ulfsproject.zefhub.io',    
    }

    transact_wish = ET.ZefFXService(**fields_d) | g_process 
    return transact_wish


def create_http_server(eff: Dict, server_zr: ZefRef) -> Dict:
    from ...core.fx.http import OurHTTPServer, Handler, _effects_processes
    from ...core.logger import log
    from ...core.op_structs import Awaitable
    import threading

    # create pushable stream and zef local variables
    server_uuid = None
    pushable_stream = None
    server = None
    thread = None

    try:
        open_requests = {}

        # Retrieve the pushable stream from the graph
        pushable_stream = Awaitable(to_ezefref(server_zr | Out[RT.PushableStream] | collect), True)
        server_uuid = str(uid(server_zr))

        # TODO Get this information from the graph
        if eff.get('pipe_into', None) is not None:  pushable_stream | eff['pipe_into']

        port = eff.get('port', 5000)
        bind_address = eff.get('bind_address', "localhost")
        do_logging = eff.get('logging', True)

        # Create the context for this server
        zef_locals = {
            "open_requests": open_requests,
            "stream" : pushable_stream,
            "server_uuid": server_uuid,
            "port": port
        }
        # Instantiate HTTP server
        server = OurHTTPServer((bind_address, port), Handler, do_logging=do_logging)

        # Set Zef local variables
        zef_locals["server"] = server
        server.zef = zef_locals


        # Start HTTP server in a thread
        thread = threading.Thread(target=server.serve_forever, daemon=True)
        zef_locals["thread"] = thread

        _effects_processes[server_uuid] = zef_locals

        thread.start()

        log.debug(f"http_server started", uuid=server_uuid, port=port, bind_address=bind_address)
        return {"server_uuid": server_uuid}
    except Exception as exc:
        raise RuntimeError(f'Error starting HTTP server: in FX.HTTP.StartServer handler: {exc}')


fx_dispatch_dict = {
    FX.HTTP.StartServer: {"create_http_server": create_http_server, "create_eff_to_process_graph_wish": create_eff_to_process_graph_wish},
}

def fx_runtime(eff: Dict):
    from ...core.fx import _effect_handlers
    if not isinstance(eff, Dict): raise TypeError(f"run(x) called with invalid type for x: {type(eff)}. You can only run a wish, which are represented as dictionaries.")
    
    if eff['type'] in fx_dispatch_dict:

        if eff['type'] == FX.HTTP.StartServer:
            http_server_d = fx_dispatch_dict[FX.HTTP.StartServer]
            create_eff_to_process_graph_wish = http_server_d['create_eff_to_process_graph_wish']
            create_http_server = http_server_d['create_http_server']

            # Stage 1
            transact_wish = create_eff_to_process_graph_wish(eff)
            server_zr = _effect_handlers[FX.Graph.Transact.d](transact_wish)

            # Stage 2
            result_d = create_http_server(eff, server_zr)
            return result_d
    
    
    handler = _effect_handlers[eff['type'].d]
    return handler(eff)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import report_import
report_import("zef.core")

####################################
# * Locating libzef
#----------------------------------

import ctypes, sys, sysconfig, os

# ** Circular import checks
import os
if os.environ.get("ZEFDB_DEVELOPER_CIRCULAR_IMPORTS", "FALSE") == "TRUE":
    from .circular_imports import check_circular_imports
    check_circular_imports()

# ** TLS certificates
# See comments in libzef CMakeLists.txt. This is to pass appropriate certificate
# information to libzef if it is bundled with pyzef.
try:
    # I don't think this will ever fail as python requires an SSL to build now I
    # think. What this block should really be doing is checking for OpenSSL not
    # LibreSSL... or maybe the libzef library should be using LibreSSL on macos.
    import ssl
except ImportError:
    raise ImportError("zef requires OpenSSL to be installed.")

_ssl_paths = ssl.get_default_verify_paths()
os.environ["LIBZEF_CA_BUNDLE"] = _ssl_paths.cafile or ""
os.environ["LIBZEF_CA_PATH"] = _ssl_paths.capath or ""

# ** auth.html location for bundled installs
# This is only relevant for windows, but we can set it regardless
import pathlib
os.environ["LIBZEF_AUTH_HTML_PATH"] = str(pathlib.Path(__file__).resolve().parent.parent)

########################################################
# * Exposing common functions
#------------------------------------------------------
from .. import pyzef
from ._core import *

# Force patching to execute
from . import patching
# Also override merge
from .overrides import *

# This set of imports is to define the order. Later imports are the ones to
# actually provide useful exports.
from . import internals
from . import pure_utils
from . import VT
from . import _error
from . import generators
from .VT import extended_containers
from . import user_value_type
from . import _image
from . import _decimal
from . import _bytes
# Up to here, DEFINITELY no zefops can be called
from . import abstract_raes
from . import graph_slice
from . import delegates
from . import op_structs
from . import _ops
from . import zef_functions
from . import graph_delta
from . import flat_graph
from . import fx
from . import serialization
from . import graph_events
from . import streams
from . import z_expression

from .VT import *

from .fx import FX

from .units import unit

from .zef_functions import func

from .serialization import serialize, deserialize

from .symbolic_expression import SV, SVs, V
from .z_expression import ZZ    # TODO: rename this to "Z" and replace the ZefOp "Z"

from .graph_events import assigned, terminated, instantiated

# Implementations come last, so that they can make use of everything else
from . import op_implementations


# Because I can't figure out why vscode displaying rich console output removes
# colours, even after forcing them in the string output, we have to hook into
# the ipython excepthook to print directly to the console.
#
# This is not needed for an ipython shell, hence the _is_jupyter check.
def visual_exception_view(error_value):
    from zef.core._error import zef_ui_err
    try:
        # zef_ui_err(error_value.wrapped)
        print(error_value.wrapped)
    except Exception as e:
        try:
            e_s = str(e)
        except:
            e_s = "Can't take str of failure exception"
        print("Failed in displaying zef error: {e_s}")
        pass
try:
    from IPython import get_ipython
    ip = get_ipython()
    # Use the same check as what rich does
    import rich.console
    if rich.console._is_jupyter():
        def ip_exception_handler(self, etype, evalue, tb, tb_offset=None):
            from ._error import ExceptionWrapper
            if etype == ExceptionWrapper:
                # Replace the wrapper object so that we don't output twice
                self.showtraceback((etype, "see visual below", tb), tb_offset=tb_offset)  # standard IPython's printout
                # Show our fancy view
                visual_exception_view(evalue)
            else:
                return self.showtraceback((etype, evalue, tb), tb_offset=tb_offset)  # standard IPython's printout

        # Overloading ipython exception handler
        ip.set_custom_exc((Exception,), ip_exception_handler) 
except:
    pass

pyzef.internals.finished_loading_python_core()

# ############################################
# # * Starting the butler
# #------------------------------------------

# # Putting this into a function to avoid poluting namespace
# def _autostart_behaviour():
#     auto_start = ops.config("butler.autoStart", KW.get)
#     if auto_start == "true":
#         internals.initialise_butler()

# _autostart_behaviour()
        



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import report_import
report_import("zef.core.user_value_type")

#                           _   _                  __     __        _               _____                                             
#                          | | | | ___   ___  _ __ \ \   / /  __ _ | | _   _   ___ |_   _| _   _  _ __    ___                         
#   _____  _____  _____    | | | |/ __| / _ \| '__| \ \ / /  / _` || || | | | / _ \  | |  | | | || '_ \  / _ \    _____  _____  _____ 
#  |_____||_____||_____|   | |_| |\__ \|  __/| |     \ V /  | (_| || || |_| ||  __/  | |  | |_| || |_) ||  __/   |_____||_____||_____|
#                           \___/ |___/ \___||_|      \_/    \__,_||_| \__,_| \___|  |_|   \__, || .__/  \___|                        
#                                                                                          |___/ |_|                                  

from .VT import String, ValueType, Dict, Any, make_VT, is_type_name_, PyDict

_user_value_type_registry = {}    # append to this: keep track of all types known to this runtime

# class UserValueType_:
#     def __init__(self, name: String, representation_type, constraints):
#         from random import randint
#         self.name = name
#         self.representation_type = representation_type
#         self.constraints = constraints
#         self.user_type_id = str(randint(0, 100000000000))     # TODO: use a Ref: the type is originally created on the respective process graph

#         # add to registry
#         _user_value_type_registry[self.user_type_id] = self

#     def __repr__(self):
#         return f"UserValueType(name={self.name}, representation_type={self.representation_type}, constraints={self.constraints})"

#     def __call__(self, *args, **kwargs):
#         """
#         equivalent to the constructor for this custom type
#         """
        
#         if bool(kwargs):
#             # if keyword args, then the representation_type MUST be a dict and no positional args are allowed
#             assert args == ()
#             assert self.representation_type in {Dict, dict}
#             val = kwargs

#         else:
#             # if positional args, then the representation_type MUST be a tuple and no keyword args are allowed
#             assert bool(kwargs)==False    # no keyword args
#             if len(args) == 1:         # if there is only one positional arg, then it is the value itself
#                 val = args[0]
#             elif len(args) == 0 and self.representation_type in {Dict, dict}:
#                 # The constructor is valid without args only in the case of a dictionary
#                 val = {}
#             else:
#                 raise ValueError("Error initializing UserValueType")

#         try:
#             cast_val = self.representation_type(val)
#         except:
#             raise Exception(f"UserValueType(name={self.name}) cannot cast '{val}' to {self.representation_type}")

#         # check whether the constraints are satisfied
#         if not is_a(cast_val, self.constraints):
#             raise Exception(f"UserValueType(name={self.name}) constraint does not match")
        
#         return UserValueInstance_(self.user_type_id, cast_val)
        
#     def __eq__(self, other):
#         if type(other) is not type(self):
#             return False
#         if other.user_type_id != self.user_type_id:
#             return False
#         return True

#     # TODO: just change this to a ValueType directly instead of this hack
#     def __instancecheck__(self, other):
#         if not isinstance(other, UserValueInstance_):
#             return False
#         return other.user_type_id == self.user_type_id
#def UVT_ctor(self, name: String, representation_type, constraints):
def UVT_ctor(self, *args, **kwargs):
    if "user_type_id" in self._d:
        # if bool(kwargs):
        #     # if keyword args, then the representation_type MUST be a dict and no positional args are allowed
        #     assert args == ()
        #     assert self.representation_type in {Dict, dict}
        #     val = kwargs

        # else:
        #     # if positional args, then the representation_type MUST be a tuple and no keyword args are allowed
        #     assert bool(kwargs)==False    # no keyword args
        #     if len(args) == 1:         # if there is only one positional arg, then it is the value itself
        #         val = args[0]
        #     elif len(args) == 0 and self.representation_type in {Dict, dict}:
        #         # The constructor is valid without args only in the case of a dictionary
        #         val = {}
        #     else:
        #         raise ValueError("Error initializing UserValueType")
        # try:
        #     cast_val = self.representation_type(val)
        # except:
        #     raise Exception(f"UserValueType(name={self.name}) cannot cast '{val}' to {self.representation_type}")

        try:
            cast_val = self._d["representation_type"](*args, **kwargs)
        except:
            raise ValueError("Couldn't construct type")

        # check whether the constraints are satisfied
        if not isinstance(cast_val, self._d["constraints"]):
            raise Exception(f"UserValueType(name={self._d['name']}) constraint ({self._d['constraints']}) does not match")
        
        return UserValueInstance(self._d["user_type_id"], cast_val)
    else:
        name, representation_type, constraints = args
        allowed_keys = {"forced_uid", "object_methods"}
        assert all(x in allowed_keys for x in kwargs)
        the_uid = kwargs.get("forced_uid", None)
        if the_uid is None:
            from random import randint
            the_uid = str(randint(0, 100000000000))
        object_methods = kwargs.get("object_methods", {})
        new_uvt = self._replace(
            name=name,
            representation_type=representation_type,
            constraints=constraints,
            user_type_id=the_uid,     # TODO: use a Ref: the type is originally created on the respective process graph
            object_methods=object_methods,
        )

        # add to registry
        assert new_uvt._d["user_type_id"] not in _user_value_type_registry
        _user_value_type_registry[new_uvt._d["user_type_id"]] = new_uvt
        return new_uvt

def UVT_str(self):
    if "name" in self._d:
        return f"UserValueType(name={self._d['name']}, representation_type={self._d['representation_type']}, constraints={self._d['constraints']})"
    else:
        return "UserValueType"

        
def UVT_is_a(x, typ):
    if "user_type_id" in typ._d:
        if not isinstance(x, UserValueInstance):
            return False
        return x._user_type_id == typ._d["user_type_id"]
    else:
        if isinstance(x, UserValueInstance):
            # Or should this be false?
            return True
        if is_type_name_(x, "UserValueType"):
            return True
        return False

UserValueType = make_VT('UserValueType',
                        constructor_func=UVT_ctor,
                        pass_self=True,
                        str_func=UVT_str,
                        is_a_func=UVT_is_a)

class UserValueInstance_:
    def __init__(self, user_type_id, value):
        # only store the id here, not the name?
        self._user_type_id = user_type_id
        self._value = value

    def _get_type(self):
        return _user_value_type_registry[self._user_type_id]
        
    def __repr__(self):
        # look up the name in the _user_value_type_registry from the id
        type_name = self._get_type()._d["name"]
        return f"{type_name}({repr(self._value)})"

    def __eq__(self, other):
        if type(other) is not type(self):
            return False
        if other._user_type_id != self._user_type_id:
            return False
        if other._value != self._value:
            return False
        return True

    def __hash__(self):
        from .VT.value_type import hash_frozen
        return hash_frozen((self._user_type_id, self._value))
    

    def __getattr__(self, other):
        if other.startswith("_"):
            return object.__getattribute__(self, other)
        typ = self._get_type()
        getattr_func = typ._d["object_methods"].get("getattr", None)
        if getattr_func is not None:
            return getattr_func(self, other)
        if isinstance(self._value, PyDict):
            return self._value[other]
        else:
            return getattr(self._value, other)

    def __dir__(self):
        if isinstance(self._value, PyDict):
            return list(self._value.keys())
        return dir(self._value)

    # required to enable dict(my_user_value_type_instance)
    # fingers crossed that nobody has a field 'keys' 
    # and tries to do my_user_value_type_instance.keys
    def _keys(self):
        if not isinstance(self._value, PyDict):
            raise AttributeError("UserValueInstance 'keys' only works on dicts")
        return self._value.keys()

    def __getitem__(self, key):
        if not isinstance(self._value, PyDict):
            raise AttributeError("UserValueInstance 'getitem' only works on dicts")
        return self._value[key]

    def __contains__(self, key):
        if not isinstance(self._value, PyDict):
            raise AttributeError("UserValueInstance 'contains' only works on dicts")
        return key in self._value

UserValueInstance = make_VT('UserValueInstance', pytype=UserValueInstance_)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# from .value_type import *
from . import ValueType, make_VT, PyDict, PyList, PyTuple
from .helpers import remove_names, absorbed, type_name

def make_distinct(v):
    """
    utility function to replace the 'distinct'
    zefop and preserve order.
    """
    seen = set()
    for el in v:
        if isinstance(el, set):
            # we can't import the following at the top of file or even outside this if
            # block: this function executes upon import of zef and leads to 
            # circular dependencies. We do really want to use the value_hash zefop though.
            #
            # TODO: Move this to pure_utils
            from ..op_implementations.implementation_typing_functions import value_hash
            h = value_hash(el)
        else:
            h = el
        if h not in seen:
            yield el
            seen.add(h)

def is_union_VT(y):
    return isinstance(y, ValueType) and y._d['type_name'] == 'Union'
def is_intersection_VT(y):
    return isinstance(y, ValueType) and y._d['type_name'] == 'Intersection'


def get_union_intersection_subtypes(typ):
    subtypes = remove_names(absorbed(typ))[0]
    if not isinstance(subtypes, PyTuple):
        subtypes = (subtypes,)
    return subtypes


def union_validation(typ):
    abs = remove_names(absorbed(typ))
    if len(abs) != 1:
        raise Exception(f"Union needs exactly 1 absorbed type or tuple of types (has {abs})")
    subtypes = abs[0]
    if not isinstance(subtypes, PyTuple):
        subtypes = (subtypes,)
    assert all(isinstance(el, ValueType) for el in subtypes), "Union requires ValueTypes"
    return True

def union_is_a(val, typ):
    assert union_validation(typ)
    subtypes = get_union_intersection_subtypes(typ)
    return any(isinstance(val, subtyp) for subtyp in subtypes)

def union_override_subtype(union, typ):
    assert union_validation(union)
    subtypes = get_union_intersection_subtypes(union)
    return all(issubclass(x, typ) for x in subtypes)

def union_is_subtype(other, union):
    assert union_validation(union)
    subtypes = get_union_intersection_subtypes(union)
    return any(issubclass(other, x) for x in subtypes)

def union_simplify(x):
    """
    Only simplifies nested Union and Intersection.
    It does not change any absorbed values, but
    only rearranges on the outer level of Zef Values.

    Union and Intersection behave analogous to + and *.

    I[U[A, B], C] = I[U[A,C], U[B,C]]              # True
    I[U[Ev, Od], Pos] = I[U[Ev,Pos], U[Odd,Pos]]   # True
    (A+B)*C = A*C + B*C
    
    U[I[Ev, Od], Pos] = U[I[Ev,Pos], I[Odd,Pos]]   # This is False!!!!


    """
    # flatten out unions: Union[Union[A][B]][C]  == Union[A][B][C]
    old_subtypes = get_union_intersection_subtypes(x)
    types = tuple(get_union_intersection_subtypes(el) if is_union_VT(el) else (el,) for el in old_subtypes)  # flatten this out
    flattened = list(make_distinct(a2 for a1 in types for a2 in a1))
    supers = []
    for i,x in enumerate(flattened):
        for j,y in enumerate(flattened):
            if j != i:
                if issubclass(x,y):
                    break
        else:
            supers.append(x)
    if len(supers) == 1:
        return supers[0]
    return Union[tuple(supers)]


Union = make_VT('Union',
                is_a_func=union_is_a,
                is_subtype_func=union_is_subtype,
                # simplify_type_func=union_simplify,
                override_subtype_func=union_override_subtype)

def intersection_validation(typ):
    abs = remove_names(absorbed(typ))
    if len(abs) != 1:
        raise Exception("Intersection needs exactly 1 absorbed type or tuple of types")
    subtypes = abs[0]
    if not isinstance(subtypes, PyTuple):
        subtypes = (subtypes,)
    assert all(isinstance(el, ValueType) for el in subtypes), "Intersection requires ValueTypes"
    return True

def intersection_is_a(val, typ):
    assert intersection_validation(typ)
    subtypes = get_union_intersection_subtypes(typ)
    return all(isinstance(val, subtyp) for subtyp in subtypes)

warned_about_intersection = False
def intersection_override_subtype(intersection, typ):
    assert intersection_validation(intersection)
    subtypes = get_union_intersection_subtypes(intersection)
    # raise NotImplementedError("This is tricky!")

    # An intersection X is a subtype of Y, if the intersection of X and ~Y is
    # empty. Since we can't prove this yet, go for a cop-out.
    #
    # For now do the easy cases and return "maybe" otherwise
    result = any(issubclass(x, typ) for x in subtypes)
    if result is False:
        global warned_about_intersection
        if not warned_about_intersection:
            print("WARNING: using intersection subtype override is not rigorous yet")
            print("WARNING: using intersection subtype override is not rigorous yet")
            print("WARNING: using intersection subtype override is not rigorous yet")
            print("WARNING: using intersection subtype override is not rigorous yet")
            print("WARNING: using intersection subtype override is not rigorous yet")
            warned_about_intersection = True

        result = "maybe"
    return result


def intersection_simplify(x):
    """
    Only simplifies nested Union and Intersection.
    It does not change any absorbed values, but
    only rearranges on the outer level of Zef Values.

    Union and Intersection behave analogous to + and *.

    I[U[A, B], C] = I[U[A,C], U[B,C]]              # True
    I[U[Ev, Od], Pos] = I[U[Ev,Pos], U[Odd,Pos]]   # True
    (A+B)*C = A*C + B*C
    
    U[I[Ev, Od], Pos] = U[I[Ev,Pos], I[Odd,Pos]]   # This is False!!!!


    """
    # flatten out Intersections: Intersection[Intersection[A][B]][C]  == Intersection[A][B][C]
    old_subtypes = get_union_intersection_subtypes(x)
    types = tuple(get_union_intersection_subtypes(el) if is_intersection_VT(el) else (el,) for el in old_subtypes)  # flatten this out
    flattened = list(make_distinct(a2 for a1 in types for a2 in a1))
    supers = []
    for i,x in enumerate(flattened):
        for j,y in enumerate(flattened):
            if j != i:
                if issubclass(y,x):
                    break
        else:
            supers.append(x)

    if len(supers) == 1:
        return supers[0]
    # TODO: If there is no overlap between any of the individual sets, this should return the empty set
    return Intersection[tuple(supers)]

Intersection = make_VT('Intersection',
                       is_a_func=intersection_is_a,
                       # simplify_type_func=intersection_simplify,
                       override_subtype_func=intersection_override_subtype)

def complement_validation(typ):
    abs = remove_names(absorbed(typ))
    if len(abs) != 1:
        raise Exception(f"Complement requires a single type not {abs}")
    assert isinstance(abs[0], ValueType), f"Complement require a ValueType not {abs[0]}"

def complement_is_a(val, typ):
    complement_validation(typ)
    subtype = remove_names(absorbed(typ))[0]
    return not isinstance(val, subtype)


make_VT('Complement',
        is_a_func=complement_is_a)

def is_validation(typ):
    abs = remove_names(absorbed(typ))
    if len(abs) != 1:
        raise Exception(f"Is requires a single predicate not {abs}")

    subtype = abs[0]
    
    # TODO: I want to change this to just "callables" as otherwise the different
    # behaviour could lead to unexpected problems
    from typing import Callable
    from ..op_structs import ZefOp
    from .. import func
    if isinstance(subtype, (ZefOp, Callable)):
        return True
    raise Exception(f'"Is[...]" called with unsupported type {subtype}')

def is_is_a(el, typ):
    is_validation(typ)
    predicate = remove_names(absorbed(typ))[0]
    try:
        if not predicate(el): return False
    except:
        return False
    return True

make_VT('Is',
        is_a_func=is_is_a)

def setof_ctor(self, *args):
    """
    Can be called with either SetOf[5,6,7] or SetOf(5,6,7).
    When calling with square brackets, a tuple must always be 
    passed implicitly or explicitly.
    SetOf[42] is NOT valid
    SetOf[(42,)] is valid
    SetOf[42,] is valid    
    SetOf[42, 43] is valid    
    """
    abs = remove_names(absorbed(self))
    if len(abs) > 0:
        return NotImplemented
    return self[args]

def setof_validation(typ):
    assert type_name(typ) == "SetOf"
    items = remove_names(absorbed(typ))
    if len(items) >= 2:
        raise Exception("SetOf can't have multiple absorbed arguments")
    return True

def setof_get_items(typ):
    setof_validation(typ)
    items = remove_names(absorbed(typ))
    if len(items) == 0:
        return ()
    return items[0]

def setof_is_a(x, typ):
    abs = setof_get_items(typ)
    return x in abs

def setof_override_subtype(setof, typ):
    return all(isinstance(x, typ) for x in setof_get_items(setof))

make_VT('SetOf',
        constructor_func=setof_ctor,
        pass_self=True,
        is_a_func=setof_is_a,
        override_subtype_func=setof_override_subtype)



def pattern_validation(self):
    abs = remove_names(absorbed(self))
    if len(abs) != 1:
        raise Exception(f"Pattern requires a single absorbed item not {abs}")
        
    pattern = abs[0]
    if not isinstance(pattern, PyDict | PyList):
        raise Exception(f"Pattern takes either a dictionary or a list not {pattern}")

    return True

def pattern_vt_matching(x, typ):
    assert pattern_validation(typ)
    p = remove_names(absorbed(typ))[0]

    class Sentinel: pass
    sentinel = Sentinel() 

    # Note: by this point, we only have access to PyDict and not Dict
    if not ((isinstance(x, PyDict) and isinstance(p, PyDict)) or
            (isinstance(x, PyList) and isinstance(p, PyList))):
        return False
    if isinstance(x, PyDict):
        for k, v in p.items():            
            r = x.get(k, sentinel)
            if r is sentinel: return False
            if not isinstance(v, ValueType): raise ValueError(f"The pattern passed didn't have a ValueType but rather {v}")
            if not isinstance(r, v): return False  
        return True
    elif isinstance(x, PyList):
        for p_e, x_e in zip(p, x): # Creates tuples of pairwise elements from both lists
            if not isinstance(p_e, ValueType):
                raise ValueError(f"The pattern passed didn't have a ValueType but rather {p_e}")
            if not isinstance(x_e, p_e): return False  
        return True

    raise NotImplementedError(f"Pattern ValueType isn't implemented for {x}")


make_VT('Pattern', is_a_func=pattern_vt_matching)




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
In Zef the types of Zef Values are values themselves: ValueType_.
These are distinct from the host language's implementation type
of the value instance.

In contrast to most programming languages's type system, the 
Zef Type system is based on set theory and has a stronger focus 
on succinctly being able to express complex domains and associative
ontolgies.

ValueTypes fall into the category of Zef Values that can absorb other
Zef Values, which is useful in constructing query expressions (as native
data structures) and graph commands. 
It also allows them to be flexibly used by the user to build custom DSLs.

---------------------- internal structure ---------------------------
It also has "absorbed": a tuple of other values


----------- Nesting Order ----------------
TODO: Do we want to keep this? Probably not.


What is "Union"? A basic value type.
Union[Int, Union[Float, String]]

Union[Int,Float]
Union[Int,Float,String]

This can be distinguished from Tuples as a type
Union[(Int,Float,String),]


----------- Constructor Overloading ----------------
Int(3.4) can have overloaded behavior. The user 
provides a function.

------------ Square Bracket Overloading --------
Should Union be of type ZefType? Or just return a ZefType?


Intersection
Union
Is

There is also a Union ValueType_, which is returned.
This custom structure may rearrange and parse terms.


---------- User Defined Types ----------
needed?
"""
# from .._error import Error
from .._core import *
from .. import internals


# For certain value types, the user may want to call e.g. Float(4.5).
# This looks up the associated function 
_value_type_constructor_funcs = {}
_value_type_attr_funcs = {}
_value_type_is_a_funcs = {}
_value_type_is_subtype_funcs = {}
# These funcs are for the VERY limited group that combines sets together.
_value_type_override_subtype_funcs = {}
_value_type_simplify_type_funcs = {}
_value_type_str_funcs = {}
_value_type_pytypes = {}

class ValueType_:
    """ 
    Zef ValueTypes are Values themselves.
    """
    def __init__(self, type_name:str, absorbed=(), pytype=None, constructor_func=None, pass_self=False, attr_funcs=(None,None,None), is_a_func=None, is_subtype_func=None, override_subtype_func=None, simplify_type_func=None, str_func=None):
            self._d = {
                'type_name': type_name,
                'absorbed': absorbed,
                'alias': None,
            }

            if constructor_func is not None:
                assert type_name not in _value_type_constructor_funcs
                _value_type_constructor_funcs[type_name] = (constructor_func, pass_self)
            if is_a_func is not None:
                assert type_name not in _value_type_is_a_funcs
                _value_type_is_a_funcs[type_name] = is_a_func
            if is_subtype_func is not None:
                assert type_name not in _value_type_is_subtype_funcs
                _value_type_is_subtype_funcs[type_name] = is_subtype_func
            if override_subtype_func is not None:
                assert type_name not in _value_type_override_subtype_funcs
                _value_type_override_subtype_funcs[type_name] = override_subtype_func
            if str_func is not None:
                assert type_name not in _value_type_str_funcs
                _value_type_str_funcs[type_name] = str_func
            if simplify_type_func is not None:
                assert type_name not in _value_type_simplify_type_funcs
                _value_type_simplify_type_funcs[type_name] = simplify_type_func
            if pytype is not None:
                assert type_name not in _value_type_pytypes
                _value_type_pytypes[type_name] = pytype
            if attr_funcs != (None, None, None):
                assert type_name not in _value_type_attr_funcs
                _value_type_attr_funcs[type_name] = attr_funcs

    def _replace(self, **kwargs):
        new_vt = ValueType_(self._d["type_name"])
        new_vt._d = dict(self._d)
        new_vt._d.update(kwargs)
        return new_vt


    def __instancecheck__(self, instance):
        return is_a_(instance, self)
    def __subclasscheck__(self, subclass):
        return is_subtype_(subclass, self) is True

    # def __repr__(self):
    def __get_nice_name(self):
        if self._d['alias'] != None:
            return self._d['alias']
        return self._d['type_name']

    def __str__(self):
        str_func = _value_type_str_funcs.get(self._d["type_name"], None)
        if str_func is None:
            if self._d["alias"] is not None:
                return self._d["alias"]
            return self.__get_nice_name() + (
                ''.join([ f"[{repr(el)}]" for el in self._d['absorbed'] ])
            )
        return str_func(self)

    def __repr__(self):
        return str(self)


    def __call__(self, *args, **kwargs):
        try:
            f,pass_self = _value_type_constructor_funcs[self._d["type_name"]]
        except KeyError:
            try:
                f = _value_type_pytypes[self._d["type_name"]]
                pass_self = False
            except KeyError:
                from .. import Error
                raise Exception(f'{self.__get_nice_name()}(...) was called, but no constructor function was registered for this type')
                return Error(f'{self.__get_nice_name()}(...) was called, but no constructor function was registered for this type')
        if pass_self:
            out = f(self, *args, **kwargs)
        else:
            out = f(*args, **kwargs)
        if out is NotImplemented:
            raise Exception("Constructor cannot be called for this instance of value type {self.__get_nice_name()}")
        return out


    def __getitem__(self, x):
        new_absorbed = self._d["absorbed"] + (x,)
        return self._replace(absorbed=new_absorbed)

    def __eq__(self, other):
        if isinstance(other, type) and other in [internals.ZefRef, internals.EZefRef]:
            import traceback
            traceback.print_stack()
            print(f"""Warning, you tried to compare == between a '{other}' and a ValueType. This is likely because you wrote `type(x) == SomeValueType`. Instead you should write `isinstance(x, SomeValueType)` or `representation_type(x) == SomeValueType`.""")
        if not isinstance(other, ValueType_): return False
        # return self._d['type_name'] == other._d['type_name'] and self._d['absorbed'] == other._d['absorbed']
        return self._d == other._d


    def __hash__(self):
        # return hash(self._d['type_name']) ^ hash(self._d['absorbed'])
        return hash_frozen(self._d)


    def __or__(self, other):
        from . import Union
        if isinstance(other, ValueType_):
            return simplify_type(Union[self, other])
        else:
            return NotImplemented

    def __ror__(self, other):
        # Is | commutative? Going to assume it isn't
        if isinstance(other, ValueType_):
            from . import Union
            return simplify_type(Union[other, self])
        else:
            return NotImplemented
    
    def __and__(self, other):
        if isinstance(other, ValueType_):
            from . import Intersection
            return simplify_type(Intersection[self, other])
        else:
            return NotImplemented

    def __rand__(self, other):
        # Is & commutative? Going to assume it isn't
        if isinstance(other, ValueType_):
            from . import Intersection
            return simplify_type(Intersection[other, self])
        else:
            return NotImplemented

    def __invert__(self):
        from . import Complement
        return Complement[self]

    def __contains__(self, x):
        """
        Allows checking membership in form
        >>> 4 in Int
        """
        return isinstance(x, self)

    def __getattribute__(self, name):
        if name.startswith("_"):
            return object.__getattribute__(self, name)

        get_attr_func, set_attr_func, dir_func = _value_type_attr_funcs.get(self._d["type_name"], (None,None,None))
        if get_attr_func is None:
            raise AttributeError(name)
        return get_attr_func(self, name)

    def __setattribute__(self, name, value):
        if name.startswith("_"):
            return object.__getattribute__(self, name)

        get_attr_func, set_attr_func, dir_func = _value_type_attr_funcs.get(self._d["type_name"], (None,None,None))
        if set_attr_func is None:
            raise AttributeError
        return set_attr_func(self, name, value)

    def __dir__(self):
        get_attr_func, set_attr_func, dir_func = _value_type_attr_funcs.get(self._d["type_name"], (None,None,None))
        if get_attr_func is None:
            return object.__dir__(self)
        return dir_func(self)


def simplify_type(typ):
    if typ._d["type_name"] in _value_type_simplify_type_funcs:
        return _value_type_simplify_type_funcs[typ._d["type_name"]](typ)
    return typ
        
    
# Internal use only
def is_type_(typ):
    return type(typ) == ValueType_

def is_a_(obj, typ):
    assert is_type_(typ), f"Can't do a is_a_ on a non-ValueType '{typ}'"
    if typ._d["type_name"] in _value_type_is_a_funcs:
        out = _value_type_is_a_funcs[typ._d["type_name"]](obj, typ)
        if out is not NotImplemented:
            return out
    if typ._d["type_name"] in _value_type_pytypes:
        from .helpers import remove_names, absorbed
        # If something is absorbed, then we don't know how to handle it so error
        abs = remove_names(absorbed(typ))
        if len(abs) > 0:
            raise Exception(f"Absorbed arguments in a {type_name(typ)} cannot be used (have {abs}), as it is a native-python type domainted ValueType")
        return isinstance(obj, _value_type_pytypes[typ._d["type_name"]])
    else:
        raise Exception(f"ValueType '{typ._d['type_name']}' has no is_a implementation")

def is_subtype_(typ1, typ2):
    assert is_type_(typ1), f"is_subtype got a non-type: {typ1}"
    assert is_type_(typ2), f"is_subtype got a non-type: {typ2}"

    if typ1._d["type_name"] in _value_type_override_subtype_funcs:
        result = _value_type_override_subtype_funcs[typ1._d["type_name"]](typ1, typ2)
        if result is True or result is False:
            return result
        # Otherwise, "maybe" and continue on, in the hopes the other type knows how.

    if typ2._d["type_name"] in _value_type_is_subtype_funcs:
        return _value_type_is_subtype_funcs[typ2._d["type_name"]](typ1, typ2)

    # Fallback
    if typ1._d["type_name"] == typ2._d["type_name"]:
        # Going to default to nonvariant
        return typ1._d == typ2._d
    else:
        # raise Exception(f"ValueType '{typ1._d['type_name']}' has no is_subtype implementation")
        return False

def is_strict_subtype_(typ1, typ2):
    res = is_subtype_(typ1, typ2)

    if typ1 == typ2:
        return False
    elif res is True:
        return True
    elif res is False:
        return False
    elif res == "maybe":
        return False
    
def is_empty_(typ):
    typ = simplify_type(typ)
    from . import SetOf
    return typ == SetOf


def is_type_name_(typ, name):
    return isinstance(typ, ValueType_) and typ._d["type_name"] == name
    

def absorbed(typ):
    return typ._d["absorbed"]
def type_name(typ):
    return typ._d["type_name"]

# Temporary hash, probably needs to be merged into other code

def hash_frozen(obj):
    if type(obj) == dict:
        h = hash("dict")
        for key in sorted(obj):
            h ^= hash(key)
            h ^= hash_frozen(obj[key])
        return h
    elif type(obj) == set:
        all_hs = [hash_frozen(x) for x in obj]
        h = hash("set")
        for h_i in sorted(all_hs):
            h ^= h_i
        return h
    elif type(obj) == list:
        h = hash("list")
        for x in obj:
            h ^= hash_frozen(x)
        return h
    elif type(obj) == tuple:
        h = hash("tuple")
        for x in obj:
            h ^= hash_frozen(x)
        return h
            
    return hash(obj)




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .. import *
from .._ops import *
from ...ui import *


event = {"type": "Request", "path": "/orders", "time": "2022-11-05 14:12:03  +0800"}

def random_time():
    import datetime
    import random
    return (datetime.datetime(2022, 11, 6) + random.random() * datetime.timedelta(days=1)).strftime("%Y-%m-%d %H:%M:%S")


def make_request_event(x):
    import random
    path = ["/orders", "/customers", "/gql"]
    return  {"type": "Request", "path": path[random.randint(0,2)], "time":random_time()}

range(100) | map[make_request_event] | collect



# a ascii historgam with a frame showing web server traffic. random data
s2 = """
┌── HTTP Server Traffic ─────────────────────────────────────────┐
│                                                                │
│                                                                │
|  /orders                                                       |
|  ▅▅▅▅▄▄▆▆▆▆▁▁▅▅▄▄▅▅▆▆▄▄▅▅▅▅▃▃▂▂▃▃▆▆▃▃▆▆▅▅▆▆▃▃██▅▅              |
|  /customers                                                    |
|  ▄▄▃▃▂▂▃▃▄▄▄▄▂▂▄▄▄▄▆▆▄▄██▁▁▃▃▄▄▃▃▄▄▅▅▄▄▂▂▄▄▂▂▃▃▇▇              |
|  /gql                                                          |
|  ▂▂▃▃▄▄▃▃▁▁▅▅▂▂▂▂▅▅▆▆▅▅▄▄▅▅▅▅▂▂▃▃▄▄▇▇▅▅██▃▃▅▅▄▄▄▄              |
|  🕰️ ──┬───────────────────┬────────────────────────┬──►        |
|  0        4        8        12        16        20             |
│                                                                │
│                                                                │
└────────────────────────────────────────────────────────────────┘
"""


# {"type": "Request", "path": "/orders", "time": "2022-11-05 14:12:03"},
events = range(1000) | map[make_request_event] | collect


# TODO fix time scale
@func
def histogram(data, group_key, x_key, x_range):
    spacing = 2
    histogram_blocks = ["▁", "▂", "▃", "▄", "▅", "▆", "▇", "█"]
    t_min, t_max = 0, len(histogram_blocks) - 1

    @func
    def normalize(m,r_min,r_max,t_min,t_max):
        return round((m-r_min)/(r_max-r_min)*(t_max-t_min)+t_min)

    key_to_size_groups = data | map[x_key] | group_by[group_key ] | map[lambda t: (t[0], len(t[1]))] | collect

    sorted_groups = key_to_size_groups | sort[second] | collect
    r_min, r_max = sorted_groups[0][1], sorted_groups[-1][1]
    
    key_to_size_d = dict(key_to_size_groups)

    normalized_blocks = x_range | map[lambda i_e: key_to_size_d.get(i_e, 0) | normalize[r_min][r_max][t_min][t_max] | collect] | map[lambda idx: histogram_blocks[idx] * spacing] | collect 
    return (normalized_blocks | join | collect)

@func
def process_request_events(events):
    grouped_by_path =  dict(events | group_by[get['path']] | collect)
    x_range = [i for i in range(0, 24)] 
    
    from datetime import datetime
    group_key = lambda s: datetime.strptime(s, "%Y-%m-%d %H:%M:%S").hour

    result = grouped_by_path | items | map[lambda kv: (kv[0], kv[1] | histogram[group_key][get['time']][x_range] | collect)] | concat | collect
    result += ["🕰️ ──┬───────────────────┬────────────────────────┬──►", '        '.join([str(x_range[i]) for i in range(0, len(x_range), 4)])]
    return result


from zef.core.op_implementations.yo_ascii import make_box
result = process_request_events(events)
http_box = make_box(result, "HTTP Server Traffic")

http_box2 =Frame(Text('\n'.join(result)), title= "HTTP Server Traffic", expand=False) 


# ┌── Effect Event Log ───────────────────────────────────────────────────────────────────────────────────┐
# │                                                                                                       │
# │                                                                                                       │
# |  ⚡ FX.Clipboard.CopyFrom           -                               12:14:02 - 6 mins ago     ✅        |
# |  ⚡ FX.HTTP.StartServer             Port 5002 is blocked            13:52:02 - 6 mins ago     ❌        |
# |  ⚡ FX.LocalFile.Load               file: zeyad.txt                 14:09:02 - 6 mins ago     ✅        |
# |  ⚡ FX.GraphQL.StartPlayground      -                               14:14:02 - 6 mins ago     ✅        |
# |  ⚡ FX.Graph.Load                   Loaded: worduel/main5           14:52:02 - 6 mins ago     ✅        |
# │                                                                                                       │
# │                                                                                                       │
# └───────────────────────────────────────────────────────────────────────────────────────────────────────┘


from datetime import datetime
events = [
    {"fx_type": FX.Graph.Load, "message": "Loaded: worduel/main5", "time": "2022-11-05 14:52:02  +0800", "is_error": False},
    {"fx_type": FX.HTTP.StartServer, "message": "Port 5002 is blocked", "time": "2022-11-05 13:52:02  +0800", "is_error": True},
    {"fx_type": FX.GraphQL.StartPlayground, "message": "-", "time": "2022-11-05 14:14:02  +0800", "is_error": False},
    {"fx_type": FX.Clipboard.CopyFrom, "message": "-", "time": "2022-11-05 12:14:02  +0800", "is_error": False},
    {"fx_type": FX.LocalFile.Load, "message": "file: zeyad.txt", "time": "2022-11-05 14:09:02  +0800", "is_error": False},
]

from zef.core.op_implementations.yo_ascii import make_box, pad_side
parse_time = lambda s: datetime.strptime(s, "%Y-%m-%d %H:%M:%S %z")

@func
def fx_event_parsing(event, longest_message, longest_fx_type):
    fx_type = str(event["fx_type"])
    message = event["message"]
    time = parse_time(event["time"]).time()
    is_error = event["is_error"]
    return f"💥 {pad_side(fx_type, 0, True, longest_fx_type)} {pad_side(message, 0, True, longest_message)} {time} - 6 mins ago     {'✅' if not is_error else '❌'}"


def fx_events_processing(events):
    events = events | sort[lambda x: parse_time(x['time'])] | collect
    longest_type = max(events | map[lambda d: d | get['fx_type']| func[str] | length | collect] | collect) + 5
    longest_message = max(events | map[lambda d: d | get['message']| func[str] | length | collect] | collect) + 10
    return events | map[fx_event_parsing[longest_message][longest_type]] | collect


from zef.core.op_implementations.yo_ascii import make_box

result = fx_events_processing(events)
# print(make_box(result, "Effect Event Log"))  
fx_log_box = make_box(result, "Effect Event Log")
fx_log_box2 =Frame(Text('\n'.join(result)), title= "Effect Event Log", expand=False) 


# ┌── Switches ──────────────────────────────────────────────┐
# │                                                          │
# │                                                          │
# |  ❌  Allow new RT type definitions      🆔 7988187e      |
# |  ✅  Allow new ET type definitions      🆔 37a4a100        |
# |  ❌  Allow new EN type definitions      🆔 a156e814        |
# |  ✅  Allow new EN type definitions      🆔 8bd8f2b1        |
# |  ❌  Quiet Mode                         🆔 2a6b3029        |
# |  ❌  Performance Mode                   🆔 c83d6b97        |
# |  ❌  Benchmark Mode                     🆔 e88aeb4b        |
# |  ✅  "yo" Condensed Output              🆔 80475a76        |
# │                                                          │
# │                                                          │
# └──────────────────────────────────────────────────────────┘

from zef.core.op_implementations.yo_ascii import make_box, pad_side

switches = [
    {"name": "Allow new RT type definitions", "is_on": False},
    {"name": "Allow new ET type definitions", "is_on": True},
    {"name": "Allow new EN type definitions", "is_on": False},
    {"name": "Allow new EN type definitions", "is_on": True},
    {"name": "Quiet Mode", "is_on": False},
    {"name": "Performance Mode", "is_on": False},
    {"name": "Benchmark Mode", "is_on": False},
    {"name": '"yo" Condensed Output', "is_on": True},
]

# a random uid generator
def make_uid():
    import uuid
    return str(uuid.uuid4())[:8]

@func
def switch_parsing(switch_t, longest_name):
    name, is_on =  switch_t
    return f"{'✅' if is_on == 'yes' else '❌'}  {pad_side(name, 0, True, longest_name)} 🆔 {make_uid()}"

def switches_processing(switches_dict):
    longest_name = max(switches_dict | items | map[lambda d: d[0] | length | collect] | collect) + 5
    return switches_dict | items | map[switch_parsing[longest_name]] | collect

result = switches_processing(zwitch.as_dict())
# print(make_box(result, "Switches"))

switches_box = make_box(result, "Switches")
switches_box2 =Frame(Text('\n'.join(result)), title= "Switches", expand=False) 

db  = Graph()
worduel_db = Graph("worduel/main5")

# ┌── Loaded Graphs ────────────────────────────────────────────────────────────────────────────────────────────────────┐
# │                                                                                                                     │
# │                                                                                                                     │
# |  Name            Blob Count         Size      UID                  Last Change                    Is Primary        |
# |  db              92 blobs           0MB       7fef825b4d77f2a3     2022-11-08 17:10:03 (+0800)    ✅                 |
# |  worduel_db      215973 blobs       3MB       3c723ac84a417f8f     2022-11-08 03:10:10 (+0800)    ❌                 |
# │                                                                                                                     │
# │                                                                                                                     │
# └─────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘

from zef.core.op_implementations.yo_ascii import make_box, pad_side

@func
def generate_graph_metrics_d(graph_t):
    g_name, g = graph_t
    return {
        "local_name": g_name,
        "uid": str(g.uid),
        "last_change": str(to_ezefref(g | all[TX] | last | collect) | time | collect),
        "blob_count": f"{g.graph_data.write_head} blobs",
        "size": f"{round((g.graph_data.write_head * 16) / 1E6)}MB",
        "is_primary_instance": g.graph_data.is_primary_instance,
    }


@func
def graph_metric_parsing(g_dict, longest_name, longest_blob, longest_size, longest_uid, longest_time):
    local_name = g_dict["local_name"]
    uid = g_dict["uid"]
    last_change = g_dict["last_change"]
    blob_count = g_dict["blob_count"]
    size = g_dict["size"]
    is_primary_instance = g_dict["is_primary_instance"]
    if isinstance(is_primary_instance, (int, bool)):
        is_primary_instance = '✅' if is_primary_instance else '❌'

    return f"{pad_side(local_name, 0, True, longest_name)} {pad_side(blob_count, 0, True, longest_blob)}  {pad_side(size, 0, True, longest_size)}   {pad_side(uid, 0, True, longest_uid)}  {pad_side(last_change, 0, True, longest_time)} {is_primary_instance}"


def graphs_processing():
    graphs = globals() | items | filter[lambda kv: is_a(kv[1], Graph)] | collect
    metrics_d = graphs | map[generate_graph_metrics_d] | collect

    headers = {
        "local_name": "Name",
        "uid": "UID",
        "last_change": "Last Change",
        "blob_count": "Blob Count",
        "size": "Size",
        "is_primary_instance": "Is Primary",
    }
    metrics_d = metrics_d | prepend[headers] | collect
    longest_name = max(metrics_d | map[lambda d: d | get['local_name']| func[str] | length | collect] | collect) + 5
    longest_blob = max(metrics_d | map[lambda d: d | get['blob_count']| func[str] | length | collect] | collect) + 5
    longest_size = max(metrics_d | map[lambda d: d | get['size']| func[str] | length | collect] | collect) + 3
    longest_uid = max(metrics_d | map[lambda d: d | get['uid']| func[str] | length | collect] | collect) + 3
    longest_time = max(metrics_d | map[lambda d: d | get['last_change']| func[str] | length | collect] | collect) + 3

    return metrics_d | map[graph_metric_parsing[longest_name][longest_blob][longest_size][longest_uid][longest_time]] | collect



result = graphs_processing()
graphs_box = make_box(result, "Loaded Graphs")
graphs_box2 = Frame(Text('\n'.join(result)), title= "Loaded Graphs", expand=False) 


hstack = HStack([ switches_box, http_box])
stack  = VStack([graphs_box, fx_log_box, hstack])




Frame(stack, title=Text("Zef Dashboard", color="#ffb3b3"), expand=False, box='ascii', padding=(2,2)) | show


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .. import report_import
report_import("zef.core.error")

custom_error_handling = True
import os
if os.environ.get("ZEF_ERROR_HANDLING", "").lower() in ["0", "false", "no", "off"]:
    custom_error_handling = False
def custom_error_handling_activated():
    return custom_error_handling

# This is an exception only for our internal evaluation engine logic. This error
# should never propagate outside of the evaluation engine, and instead be
# converted to an ExceptionWrapper, which will print nicely.
class Error_(Exception):
    def __init__(self, name):
        self.name = name

    # def __set_name__(self, parent, name):
    #     self.name = name

    def __call__(self, *args):
        err = Error_(self.name)
        err.args = args
        err.contexts = []
        err.nested = None
        return err

    def __repr__(self):
        if not self.args or len(self.args) == 0: args = "()"
        elif len(self.args) == 1: args = f"({repr(self.args[0])})"
        else: args = self.args
        return f'{self.name}{args}'

    def __str__(self):
        try:
            # We add an extra newline at the start to better handle the output
            return "\n" + zef_ui_err(self, as_str=True)
        except Exception as e:
            try:
                e_s = str(e)
            except:
                e_s = "Can't take str of failure exception"
            return f"Failed in displaying zef error: {e_s}"
        

    def __eq__(self, other):
        if not isinstance(other, Error_): return False
        return self.name == other.name and self.args == other.args

    def __bool__(self):
        return False

def zef_ui_err_fallback(self):
    try:
        return "\n\nThis is a custom error output for a wrapped Zef error\n\n" + str_zef_error(self)
    except Exception as exc:
        import traceback
        traceback.print_tb(exc.__traceback__)


def zef_ui_err(err, as_str=False):
    from ..ops import contains,last, get, collect, filter, ZefOp, LazyValue
    from ..ui import Text,VStack, Frame, show, Code, to_rich_str
    from ..core.op_implementations.implementation_typing_functions import ZefGenerator

    def truncate_obj_str(obj, max_len=20):
        obj_str = str(obj)
        index_of_newline = obj_str.find('\n')
        if index_of_newline != -1: obj_str = obj_str[:index_of_newline] + "..."
        if len(obj_str) > max_len: obj_str = obj_str[:max_len] + "... "
        return obj_str

    def make_lazy_value_pretty(lzy_val: LazyValue) -> str:
        if not isinstance(lzy_val, LazyValue): return str(lzy_val)
        initial_value = truncate_obj_str(lzy_val.initial_val)
        internal_ops = lzy_val.el_ops.el_ops
        internal_ops = ZefOp(internal_ops,)
        return f"{initial_value} | {internal_ops}" 
    
    MAX_CONSOLE_WIDTH = 115
    try:
        # Required Information
        nested = err.nested
        name = err.name
        contexts = err.contexts
        frames = contexts | filter[contains["frames"]] | collect 
        if frames:
            top_frame = frames | last | get["frames"] | last | collect 
        else:
            top_frame = None
        type_check = contexts | filter[contains["type_check"]] | collect

        chains = contexts | filter[contains["chain"]] | collect

        states = contexts | filter[contains["state"]] | collect

        metadata = contexts | filter[contains["metadata"]] | collect
        if len(metadata) == 1: 
            metadata = metadata[0]['metadata']
        else:
            metadata = None


        #### Title ####
        title = Text(name, color="#FF9494", italic=True)


        stack_lst = []
        ##### Header ####
        err_type = nested.get('type', None)
        if top_frame and err_type:
            stack_lst += [Text(f"\n{nested['type']} occured in {top_frame['func_name']}", color="#189ad3")]
        elif err_type:
            stack_lst += [Text(f"\n{nested['type']}", color="#189ad3")]


        ##### Frames ####
        if top_frame:
            fname = top_frame['filename']
            try:
                truncated_filename = fname[fname.rindex("zef"):]
            except:
                truncated_filename = fname
            line_no = f":{top_frame['lineno']}" if top_frame['lineno'] else ""
            file_line = Text(f"~/{truncated_filename}{line_no}", color="#33b864")
            file_text = Text(["File ", file_line])
            stack_lst += [file_text]

        ##### Carret Highlighting ####
        if chains:
            chain_str   = make_lazy_value_pretty(chains[0]['chain'])
            arrow_str   = "--> "
            if len(chain_str) + len(arrow_str) > MAX_CONSOLE_WIDTH:
                chain_str = chain_str[:MAX_CONSOLE_WIDTH - len(arrow_str) - 5] + "... "  
            stack_lst += [Text([Text("\n--> ", color="#33b864"), chain_str])]

            def make_carrets_if_found(chain_str, failed_func, arrow_str):
                if "_implementation" in failed_func: failed_func = failed_func[:failed_func.index("_implementation")]
                if "_imp" in failed_func: failed_func = failed_func[:failed_func.index("_imp")]
                idx = chain_str.find(failed_func) 
                if idx == -1: 
                    return []
                padding = idx + len(arrow_str)
                error_carrets = f"{' ' * padding}{'^' * len(failed_func)}"
                return  [Text(error_carrets, color="#FF9494")]

            failed_func = str(ZefOp((chains[-1]['op'],),))
            stack_lst += make_carrets_if_found(chain_str, failed_func, arrow_str)

        ##### States or Context #####
        if len(states) < 1:
            chain = chains[0]
            context_header = Text("\n==Context==\n", bold=True, justify="center", italic=True)
            code_str = f"""
    chain = {make_lazy_value_pretty(chain['chain'])}
    input = {truncate_obj_str(chain['input'], 40)}
    op    = {ZefOp((chain['op'],),)}
            """
            context_code = Code(code_str, language = "python3")
            stack_lst += [context_header, context_code]
        else:
            state_header = Text("\n==States==\n", bold=True, justify="center", italic=True)
            state_stack = []

            for i, chain in enumerate(reversed(chains)):
                if isinstance(chain['input'], ZefGenerator): break
                code_str = f"""
    input = {truncate_obj_str(chain['input'], 40)}
            """
                chain_state = Code(code_str, language = "python3")
                state_frame = Frame(chain_state, title=repr(ZefOp((chain['op'],),)))
                state_stack.append(state_frame)

            stack_lst += [state_header, *state_stack]


        ##### Type Checking ####
        tc_header = Text("\n==Type Checking==\n", bold=True, justify="center", italic=True)
        tc_fail_body = Text("Failed to do Type Checking due to missing annotations", color="#FF9494", bold=True)

        if type_check:
            tc_stack = []
            for t_check in type_check:
                t_check = t_check['type_check']
                if t_check:
                    tc_fn = f"{t_check['function'].__name__}"
                    tc_body = f"Expected: {t_check['expected']['arg']} = {t_check['expected']['type']}\n"
                    tc_check = f"Type Check against `{t_check['expected']['input']}`: "
                    tc_result =  ['Failed ❌','Success ✅'][t_check['result']]
                    tc_body = Frame(Text([tc_body,tc_check, tc_result], color="#FF9494"), title=Text(str(tc_fn)))
                    tc_stack.append(tc_body)
            
            if tc_stack: stack_lst += [tc_header, *tc_stack]
            else: stack_lst += [tc_header, tc_fail_body]
        else:
            stack_lst += [tc_header, tc_fail_body]

        ##### Error Messages ####
        if nested.get('args', None):
            err_msg_header = Text("\n==Error Message==", bold=True, justify="center", italic=True)
            err_msg = Text(f"\n{','.join([str(arg) for arg in nested['args']])}", color="#FF9494")
            stack_lst += [err_msg_header, err_msg]

        stack = VStack(stack_lst, expand=True)
        frame = Frame(stack, title= title, expand=True)
        if as_str:
            temp = frame | to_rich_str
            return temp
        else:
            frame | show
            return ""
    
    except Exception as exc:
        import traceback
        print(f"!Visual Error output failed!")
        try:
            print(type(exc))
            print(exc)
            print(traceback.format_exc())
        except:
            pass
        try:
            return zef_ui_err_fallback(err)
        except:
            pass


# This class exists solely to enable printing of the _ErrorType class above in
# normal exception handling. It also allows us to identify points at which we
# have ended control in the evaluation engine, and had to throw to the external
# environment.
class ExceptionWrapper(Exception):
    def __init__(self, wrapped):
        self.wrapped = wrapped

# When to use an _ErrorType or a ExceptionWrapper?
#
# Inside of evaluation-engine control, _ErrorType should be always present. This
# means whenever _ErrorTypes are present, no more traceback information of code
# is gathered. Implementation functions take want to attach metadata and
# otherwise say "my child broke, not me" should catch all other exceptions and
# turn them into _ErrorTypes.
#
# ExceptionWrappers are on the same level as regular python exceptions, and
# should only be thrown when we are giving up control back to regular python
# execution (which could be to outermost-scope or in the execution of a user
# zef-function).
#
# The only difference between an ExceptionWrapper and an Exception is that we
# (as the evalution engine) know that we can unwrap an ExceptionWrapper as
# exc.wrapped to get an _ErrorType and continue propagate that internally if we
# want.
    


#     def __new__(cls, *args):
#         return cls.BasicError(*args)

#     def __repr__(self):
#         return f'Error'

predefined_errors = [
    "TypeError",
    "RuntimeError",
    "ValueError",
    "NotImplementedError",
    "BasicError",
    "UnexpectedError",
    "MapError",
    "MatchError",
    "Panic",
]

def error_dir(self):
    return predefined_errors
def error_getattr(self, x):
    return Error_(x)
def error_ctor(*args):
    return Error.BasicError(*args)
    

from .VT import make_VT
Error = make_VT("Error",
                pytype=Error_,
                attr_funcs=(error_getattr, None, error_dir),
                constructor_func=error_ctor)

class EvalEngineCoreError(Exception):
    def __init__(self, exc):
        self.name = "EvalEngineCoreError"
        self.nested, self.frames = convert_python_exception(exc)

    def __str__(self):
        return f"This is a failure in the core evaluation engine!"#, {self.nested['type']}: {self.nested['args']}"





# Error spec

# Error = PythonType[Exception] & Object[Pattern[{
#     # Type name
#     "name": String,
#     # User added details. For generic cases is empty (e.g. UnexpectedError()
#     "args": List[Any],
#     # Error that caused this error
#     "nested": Error | PythonExcInfo | Null,
#     # Context information, up until the nested error
#     "contexts": List[Context],
# }]]

# PythonExcInfo = Pattern[{
#     # Python exception class name
#     "type": String,
#     # Contents of the exception
#     "args": List[Any],
#     # If the python exception was nested using python raising syntax.
#     "nested": Tuple[PythonExcInfo, List[FrameInfo]],
# }]

# Context = OpContext | FramesContext | MetadataContext

# OpContext = OpChainContext & Optional[ActiveOpContext | CollectingGeneratorContext]
# OpChainContext = Pattern[{
#     "chain": LazyValue & With[CollectingOp | ForEachingOp],
# }]
# # When the evaluation engine gets an error after running a particular op's
# # implementation function.
# ActiveOpContext = Pattern[{
#     # Index of the op in the chain that is being run
#     "op_i": Int,
#     # Input to this op
#     "input": Any,
# }]
# # When the evaluation engine is collecting up a ZefGenerator at the end of the
# # chain.
# CollectingGeneratorContext = Pattern[{
#     "state": "collecting",
#     # Index of the generator
#     "val_i": Int,
# }]

# FramesContext = Pattern[{
#     "frames": List[FrameInfo]
# }]
# FrameInfo = Pattern[{
#     "lineno": Int,
#     "filename": String,
#     "func_name": String,
#     "locals": Dict[String,Any],
# }]

# # This is a generic "anyone can hook in" context. Example, used by map to add
# # the context of the last input.
# MetadataContext = Pattern[{
#     "metadata": Any,
# }]
    

# * Manipulating errors

def add_error_context(error, context):
    return prepend_error_contexts(error, [context])
def prepend_error_contexts(error, contexts):
    # from copy import copy
    err = Error_(error.name)
    err.args = error.args
    err.contexts = error.contexts
    err.nested = error.nested
    out = err
    if getattr(out, "contexts", None) is None:
        out.contexts = []

    out.contexts = [*contexts, *out.contexts]

    return out


# * Displaying errors

def str_tb_up_to_zef(tb):
    import traceback
    l = filter_tb_to_before_zef(tb)
    return traceback.format_list(l)

def str_frame_info(frames):
    if len(frames) == 0:
        return "-- EMPTY (need to hide this in printing)"
    s = []
    for frame in reversed(frames):
        s.append(f"-- Frame: in func({frame['func_name']}) in '{frame['filename']}'@{frame['lineno']}")
    return '\n'.join(s)

def str_zef_error(error):
    s = ""
    if type(error) == Error_:
        if error.nested is not None:
            s += str_zef_error(error.nested)
            s += "\n^^^^^\nCaused "

        s += error.name + ": "
        if len(error.args):
            s += str(error.args)
    elif type(error) == EvalEngineCoreError:
        s += "CORE EVALUATION ENGINE ERROR!"
        s += "CORE EVALUATION ENGINE ERROR!"
        s += "CORE EVALUATION ENGINE ERROR!"
        import traceback
        s += '\n'.join(traceback.format_exception(error))
        # s += str_python_exception_dict(exc_info)
    elif type(error) == dict:
        s += str_python_exception_dict(error)
    else:
        s += f"Don't know how to print error of type {type(error)}"
        s += str(error)

    # Now make the context/traceback up
    for context in reversed(getattr(error, "contexts", [])):
        try:
            s += '\n' + zef_error_context_str(context)
        except Exception as exc:
            s += "\ncontext couldn't be printed"
            s += " " + str(exc)

    s += "\n==============================================="
    return s

def str_truncate(s):
    # return s
    if len(s) > 30:
        s = s[:15] + " ... " + s[-15:]
    return s


def str_op_chain(op):
    from . import RT
    if op[0] == RT.Function:
        func = op[1][0][1]
        import types
        if isinstance(func, types.FunctionType):
            name = func.__name__
        else:
            name = str(func)
        return f"Function: {name}"
    elif op[0] == RT.Apply:
        return f"Apply: {str_op_chain(op[1][0].el_ops[0])}"
    else:
        try:
            return str_truncate(str(op))
        except:
            return "ERROR PRINTING OP"

def zef_error_context_str(context):
    s = ""

    if "metadata" in context:
        l = []
        for key,val in context["metadata"].items():
            l += [f"@@@ '{key}' = '{val}'"]
        s += '\n'.join(l)

    if "frames" in context:
        s += str_frame_info(context["frames"])
        return s

    if "op" in context:
        # op_s = str_truncate(str(op_chain_pretty_print(context['op'])))
        s += f" evaling op {str_op_chain(context['op'])}"
    if "input" in context:
        try:
            input_s = str_truncate(str(context['input']))
        except:
            input_s = "ERROR PRINTING INPUT"
        s += f" using input ({input_s})"
    if "chain" in context:
        try:
            chain_s = str_truncate(str(context['chain']))
        except:
            chain_s = "ERROR PRINTING CHAIN"
        s += f" of chain ({chain_s})"
    if "state" in context:
        if context["state"] == "generator":
            s += f" up to i={context['i']} in generator"
            if context['i'] > 0:
                s += f" and last output was {context['last_output']}"

    if s == "":
        s = "WHY IS THIS EMPTY???"
        s += "\n" + str(context)
    return s
    
def str_python_exception_dict(d):
    s = ""
    s += f"{d['type']} - {d['args']}"
    # str_frame_info(d['frames'])
    if "nested" in d:
        nested_d,frames = d["nested"]
        s += "\n++++" + str_python_exception_dict(nested_d)
        s += "\n++++" + str_frame_info(frames)
    return s
    

# * Utils

def filter_tb_to_before_zef(tb):
    import traceback
    # Want to avoid frames that first enter the zef codebase
    l = traceback.extract_tb(tb)
    for i,item in enumerate(l):
        if "op_structs.py" in item.filename:
            break
    l = l[:i]
    return l

def convert_python_exception(exc, filter_frames=True):
    tb = exc.__traceback__
    frames = process_python_tb(tb, filter_frames)


    out = {
        "type": str(type(exc)),
        # Might need to serialize this?
        "args": exc.args,
    }

    if exc.__cause__ is not None:
        out["nested"] = convert_python_exception(exc.__cause__)
    elif exc.__context__ is not None and not exc.__suppress_context__:
        out["nested"] = convert_python_exception(exc.__context__)

    return out, frames

def process_python_tb(tb, filter=True):
    # Not sure if ops are available here so doing this in pure python
    frames = []
    while tb is not None:
        frames.append(tb.tb_frame)
        tb = tb.tb_next
    # Get rid of the first one - at least where this is currently called from we don't
    # want to include "evaluate". In the future, maybe this will have to be a switch.
    # frames = frames[1:]
    def process_frame(frame):
        return {
            "locals": frame.f_locals,
            "lineno": frame.f_lineno,
            "filename": frame.f_code.co_filename,
            "func_name": frame.f_code.co_name,
        }

    frame_info = [process_frame(frame) for frame in frames]

    if filter:
        frame_info = [frame for frame in frame_info if "op_structs.py" not in frame["filename"]]

    return frame_info


def make_custom_error(e, err_type, err_message, context):
    py_e,frames = convert_python_exception(e)
    err_instance = err_type
    if err_message: py_e['args'] = ((err_message, ))
    err_instance.nested = py_e
    err_instance = add_error_context(err_instance, {"frames": frames})
    return add_error_context(err_instance, context)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import report_import
report_import("zef.core.serialization")

__all__ = [
    "instantiated",
    "terminated",
    "assigned",
]

from .VT import *
from .VT import make_VT, insert_VT
from .VT.helpers import remove_names, absorbed

# from .abstract_raes import Entity, AttributeEntity, Relation, TXNode, Root, make_custom_entity
from .abstract_raes import make_custom_entity
# TODO Move this somewhere sensible
def is_a_custom_entity(x, typ):
    return (isinstance(x, EntityRef) and x.d["type"] == ET.ZEF_CustomEntity)
CustomEntity = make_VT("CustomEntity", is_a_func=is_a_custom_entity)


instantiated     = make_custom_entity(name_to_display='instantiated', predetermined_uid='60252a53a03086b7')
terminated       = make_custom_entity(name_to_display='terminated', predetermined_uid='4f676154ffeb9dc8')
assigned         = make_custom_entity(name_to_display='assigned', predetermined_uid='c31287dab677f38c')

infinity           = make_custom_entity(name_to_display='infinity',    predetermined_uid='4906648460291096')
nil                = make_custom_entity(name_to_display='nil',         predetermined_uid='1654670075329719') #| register_call_handler[f1] | run[execute] | get['entity'] | collect  # TODO

def has_uid_validation(typ):
    abs = remove_names(absorbed(typ))
    if len(abs) != 1:
        raise Exception(f"HasUID should have only 1 absorbed item: {abs}")
    return True
def has_uid_is_a(x, typ):
    assert has_uid_validation(typ)
    item = remove_names(absorbed(typ))[0]
    from ._ops import uid
    return uid(x) == item
# TODO: This is "early" because it is available before the uid zefop. Change this if possible.
def early_uid(x):
    return x.d["uid"]
HasUID = make_VT("HasUID", is_a_func=has_uid_is_a)
# insert_VT("Instantiated", CustomEntity & HasUID[early_uid(instantiated)])
# insert_VT("Terminated", CustomEntity & HasUID[early_uid(terminated)])
# insert_VT("Assigned", CustomEntity & HasUID[early_uid(assigned)])

# We need to make a separated type, as the absorbed arguments are relevant to tests

def make_custom_VT(name, custom_ent):
    custom_ent_uid = early_uid(custom_ent)

    def is_a_func(x, typ):
        if not isinstance(x, CustomEntity & HasUID[custom_ent_uid]):
            return False
        # TODO: We need to test the absorbed arguments covariantly
        if len(absorbed(typ)) > 0:
            print("WARNING: Not testing the absorbed arguments of isinstance with {name}.")
        return True

    return make_VT(name, is_a_func=is_a_func)

Instantiated = make_custom_VT("Instantiated", instantiated)
Terminated = make_custom_VT("Terminated", terminated)
Assigned = make_custom_VT("Assigned", assigned)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Z expressions are useful notation for succinctly declaring sets of things.
e.g. (Z > 10) denotes the set of all things greater than 10.

This can be chained though: zef functions and all kinds of operators can be absorbed.
e.g. IsJohn = (Z.first_name | to_uppercase = 'JOHN')

Z expressions sit right at the top of the food chain when it comes to operator binding.
A Z-expression gobbles up anything to become a new Z-Expression.

So how can we use them to declare sets of things? As soon as a Z expression encounters
a binary comparison operator ('=', '>', '<', '>=', '<=', '!=') it will stop gobbling
up operators and instead become a ValueType.
"""

from . import VT
from .VT import ET, Is
from .zef_functions import func

# from .VT import FlatGraph, Pattern, Any, SetOf, Val,  RT, AET
# from ._ops import match, collect, insert, split, get, filter, map, Z

class ZExpression_:
    def __init__(self, root_node=None):
        self.root_node = root_node    # if this is None, then it is a bare variable. This also qualifies as a symbolic expression.

    def __repr__(self):
        # print(f'repr called for {self.root_node}')
        # raise NotImplementedError
        if self.root_node is None:            
            return f"ZZ"
        else:
            # return "Z Expression output: todo"#self.root_node.fg | graphviz | collect
            return str(f"ZExpression({self.root_node})")

    def __hash__(self):
        return (hash(self.root_node)+78566354676)


    def __getattr__(self, other):
        return compose_zexpr(ET.Dot, self, other)

    def __getitem__(self, other):
        return compose_zexpr(ET.GetItem, self, other)

    def __add__(self, other):
        return compose_zexpr(ET.Add, self, other)

    def __radd__(self, other):
        return compose_zexpr(ET.Add, other, self)

    def __sub__(self, other):
        return compose_zexpr(ET.Subtract, self, other)

    def __rsub__(self, other):
        return compose_zexpr(ET.Subtract, other, self) 

    def __mul__(self, other):
        return compose_zexpr(ET.Multiply, self, other)

    def __rmul__(self, other):
        return compose_zexpr(ET.Multiply, other, self)

    def __truediv__(self, other):
        return compose_zexpr(ET.Divide, self, other)
    
    def __rtruediv__(self, other):
        return compose_zexpr(ET.Divide, other, self)

    def __mod__(self, other):
        return compose_zexpr(ET.Modulo, self, other)

    def __rmod__(self, other):
        return compose_zexpr(ET.Modulo, other, self)
    
    def __pow__(self, other):
        return compose_zexpr(ET.Power, self, other)

    def __rpow__(self, other):
        return compose_zexpr(ET.Power, other, self)
    
    def __or__(self, other):
        return compose_zexpr(ET.Or, self, other)
    
    def __ror__(self, other):
        return compose_zexpr(ET.Or, other, self)
    
    def __and__(self, other):
        return compose_zexpr(ET.And, self, other)
    
    def __rand__(self, other):
        return compose_zexpr(ET.And, other, self)
    
    def __xor__(self, other):
        return compose_zexpr(ET.Xor, self, other)

    def __rxor__(self, other):
        return compose_zexpr(ET.Xor, other, self)
        

    def __eq__(self, other):
        return z_expression_to_vt(ET.Equals, self, other)

    def __req__(self, other):
        return z_expression_to_vt(ET.Equals, other, self)

    def __ne__(self, other):
        return z_expression_to_vt(ET.NotEquals, self, other)

    def __rne__(self, other):
        return z_expression_to_vt(ET.NotEquals, other, self)

    def __gt__(self, other):
        return z_expression_to_vt(ET.GreaterThan, self, other)
    
    def __rgt__(self, other):
        return z_expression_to_vt(ET.GreaterThan, other, self)
    
    def __ge__(self, other):
        return z_expression_to_vt(ET.GreaterThanOrEquals, self, other)

    def __rge__(self, other):
        return z_expression_to_vt(ET.GreaterThanOrEquals, other, self)
    
    def __lt__(self, other):
        return z_expression_to_vt(ET.LessThan, self, other)

    def __rlt__(self, other):
        return z_expression_to_vt(ET.LessThan, other, self)

    def __le__(self, other):
        return z_expression_to_vt(ET.LessThanOrEquals, self, other)
    
    def __rle__(self, other):
        return z_expression_to_vt(ET.LessThanOrEquals, other, self)
    


def unwrap_zexpr(x):
    if isinstance(x, ZExpression_):
        if x.root_node is None:
            return ET.Z
        else:
            return x.root_node
    else:
        return x


def compose_zexpr(op_type: ET, arg1, arg2):
    return ZExpression_(root_node=op_type(arg1=unwrap_zexpr(arg1), arg2=unwrap_zexpr(arg2)))
   


import operator as o
_op_translation = {
    ET.Add: o.add,
    ET.Subtract: o.sub,
    ET.Multiply: o.mul,
    ET.Divide: o.truediv,
    ET.Modulo: o.mod,
    ET.Power: o.pow,
    ET.Or: o.or_,
    ET.And: o.and_,
    ET.Xor: o.xor,

    ET.GetItem: o.getitem,
    ET.Dot: lambda x, k: x.__getattr__(k),

    ET.GreaterThan: o.gt,
    ET.GreaterThanOrEqual: o.ge,
    ET.LessThan: o.lt,
    ET.LessThanOrEqual: o.le,
    ET.Equals: o.eq,
    ET.NotEquals: o.ne,
} 


def compose_fct(op_type: ET, g1, g2):
    import zef
    res = _op_translation[op_type](g1, g2)
    if isinstance(res, zef.core.op_structs.LazyValue):
        # res2 = res()
        # print(f'Compose {op_type} with {g1} and {g2} -> {res} -> {res2}')
        return res()
    else:
        # print(f'Compose {op_type} with {g1} and {g2} -> {res}')
        return res


def eval_if_lazy(x):
    import zef
    if isinstance(x, zef.core.op_structs.LazyValue):
        return x()
    else:
        return x


def z_expression_to_vt(op_type: ET, arg1, arg2) -> VT:
    """
    This function is hit when a binary comparison operator is encountered in a ZExpression.
    It returns a value type, which wraps a predicate function.
    """
       
    def step(expr) -> func:
        import zef
        is_lambda = lambda x: callable(x) and not isinstance(x, zef.core.op_structs.ZefOp_)

        # exit early if this is the main variable (ET.Z  => identity function)
        # or a constant expression, i.e. not another nested expression in the form of an EntityValueInstance_
        if expr == ET.Z:
            return lambda x: x

        if not isinstance(expr, zef.core.patching.EntityValueInstance_):
            return expr   # normal value

        # once we're here, it must be an EntityValueInstance_: recurse into its children
        f1 = step(eval_if_lazy(expr.arg1))
        f2 = step(eval_if_lazy(expr.arg2))

        # print(f"step will dispatch on: {op_type} {f1}    {f2}  {callable(f1)} {callable(f2)}")
        # one extra lambda layer to prevent evaluation of the first term
        return {
            (False, False): lambda:           compose_fct(expr._entity_type, f1, f2),
            (True, False):  lambda: lambda x: compose_fct(expr._entity_type, f1(x), f2),
            (False, True):  lambda: lambda x: compose_fct(expr._entity_type, f1, f2(x)),
            (True, True):   lambda: lambda x: compose_fct(expr._entity_type, f1(x), f2(x)),
        }[is_lambda(f1), is_lambda(f2)]()

    a = op_type(arg1=unwrap_zexpr(arg1), arg2=unwrap_zexpr(arg2))
    # print(f"\n\n----\n{a=}")
    return Is[step(a)]
 


ZZ = ZExpression_() # TODO: rename this to "Z" once the "Z" zefop can be fully replaced with this

from .VT import make_VT
ZExpression = make_VT("ZExpression", pytype=ZExpression_)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import make_VT, Error, ZefGenerator, PyList, PySet, PyTuple, PyDict, ValueType
from .value_type import is_type_name_
from .helpers import remove_names, absorbed, generic_subtype_validate, generic_subtype_get, generic_covariant_is_subtype, type_name



# TODO: Change this to a proper isa
def tuple_validation(tup):
    items = remove_names(absorbed(tup))
    if len(items) == 0:
        return True
    if len(items) >= 2:
        raise Exception("Tuple can't have more than one curried item")
    tup_params = items[0]
    assert all(isinstance(x, ValueType) for x in tup_params)
    return True

def tuple_get_params(tup):
    assert tuple_validation(tup)
    items = remove_names(absorbed(tup))
    if len(items) == 0:
        return ()
    return items[0]
    
def tuple_override_subtype(tup, typ):
    assert tuple_validation(tup)
    if is_type_name_(typ, "List"):
        return True
    return "maybe"

def tuple_is_subtype(x, tup):
    assert tuple_validation(tup)
    if type_name(x) != "Tuple":
        # TODO: Lists etc should be allowed
        return False
    assert tuple_validation(x)
    tup_params = tuple_get_params(tup)
    x_params = tuple_get_params(x)

    return all(issubclass(a,b) is True for a,b in zip(x_params, tup_params))

make_VT('Tuple', pytype=tuple, override_subtype_func=tuple_override_subtype,
        is_subtype_func=tuple_is_subtype)

def list_is_a(x, typ):
    assert generic_subtype_validate(typ)
    import sys
    from typing import Generator
    if not isinstance(x, (PyList, PyTuple, Generator, ZefGenerator)):
        return False
    subtype = generic_subtype_get(typ)
    if subtype is None:
        return True
    if isinstance(x, (Generator, ZefGenerator)):
        raise NotImplementedError()

    return all(isinstance(item, subtype) for item in x)

make_VT('List',
        constructor_func=tuple,
        is_a_func=list_is_a,
        is_subtype_func=generic_covariant_is_subtype)

def set_is_a(x, typ):
    assert generic_subtype_validate(typ)
    import sys
    if not isinstance(x, PySet):
        return False
    subtype = generic_subtype_get(typ)
    if subtype is None:
        return True

    return all(isinstance(item, subtype) for item in x)
make_VT('Set', pytype=set, is_a_func=set_is_a,
        is_subtype_func=generic_covariant_is_subtype)

def dict_validate(typ):
    # Should validation return True/False? Or True/String? Or True and throw?
    items = remove_names(absorbed(typ))
    if len(items) == 0:
        return True
    if len(items) != 2:
        raise Exception(f"Dict validation failed: there should be two subtypes for the keys and values ({items})")
    key_type, value_type = items
    if not isinstance(key_type, ValueType):
        raise Exception(f"Dict validation failed: key type is not a ValueType ({key_type})")
    if not isinstance(value_type, ValueType):
        raise Exception(f"Dict validation failed: value type is not a ValueType ({value_type})")
    return True

def dict_is_a(x, typ):
    assert dict_validate(typ)
    import sys
    if not isinstance(x, PyDict):
        return False
    abs = remove_names(absorbed(typ))
    if len(abs) == 0:
        return True

    T1, T2 = abs
    return all(isinstance(key, T1) and isinstance(val, T2) for key,val in x.items())
make_VT('Dict', pytype=dict, is_a_func=dict_is_a,
        is_subtype_func=generic_covariant_is_subtype)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .VT import *
from ._error import *
class ZefGenerator_:
    """
    A class that acts as a uniform interface instead of Python's
    builtin generator. Defining a regular generator using yield
    in Python does not always obey value semantics. It still keeps
    state when iterated over.
    
    Example:
    >>> def repeat(x, n=None):    
    >>>  def make_wrapper():
    >>>      if n is None:
    >>>          while True:
    >>>              yield x
    >>>      else:
    >>>             for _ in range(n):
    >>>                 yield x
    >>>  
    >>>     return ZefGenerator(make_wrapper)

    Returning `wrapper()` directly is equivalent to the class
    where `__iter__` returns self, thus maintaining state from
    previous iterations.
    """
    def __init__(self, generator_fct):
        self.generator_fct = generator_fct
        self.contexts = []

    def __iter__(self):
        from .op_implementations.implementation_typing_functions import wrap_error_raising
        # return self.generator_fct()
        def wrap_errors():
            it = iter(self.generator_fct())
            i = 0
            last_output = None
            while True:
                cur_context = [*self.contexts,
                               {"state": "generator",
                                "i": i,
                                "last_output": last_output}]
                try:
                    item = next(it)
                    yield item
                    i += 1
                    last_output = item
                except StopIteration:
                    return
                except GeneratorExit:
                    raise
                except Exception as e:
                    if not custom_error_handling_activated():
                        raise
                    elif isinstance(e, EvalEngineCoreError):
                        wrap_error_raising(e, cur_context)
                    elif isinstance(e, ExceptionWrapper):
                        wrap_error_raising(e, cur_context)
                    elif isinstance(e, Error_):
                        wrap_error_raising(e, cur_context)
                    else:
                        wrap_error_raising(e, cur_context)
        return wrap_errors()

    def __str__(self):
        return "ZefGenerator"

    def add_context(self, context):
        new_gen = ZefGenerator_(self.generator_fct)
        new_gen.contexts = [context, *self.contexts]
        return new_gen

from .VT import make_VT
make_VT("ZefGenerator", pytype=ZefGenerator_)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.VT")

from functools import partial
from .value_type import ValueType_, is_type_name_
from .. import internals
from ... import pyzef

__all__ = []

def insert_VT(name, vt):
    assert name not in globals()
    if vt._d["type_name"] != name:
        vt = vt._replace(alias=name)
    globals()[name] = vt
    global __all__
    __all__.append(name)
    return vt

def make_VT(name, **kwargs):
    vt = ValueType_(type_name=name, **kwargs)
    return insert_VT(name, vt)


make_VT("ValueType", pytype=ValueType_)

# I think these should become ValueTypeParmeters instead
# T          = ValueType_(type_name='T',  constructor_func=None)
# T1         = ValueType_(type_name='T1',  constructor_func=None)
# T2         = ValueType_(type_name='T2',  constructor_func=None)
# T3         = ValueType_(type_name='T3',  constructor_func=None)


# Bootstrap types
from . import python_types

# TODO: This should be replaced with the proper symbolic variables
Variable = String

from . import sets

from . import libzef_types

from . import rae_types

from . import external

from . import later

from . import codebase_query
# extended_containers relies on Error and ZefGenerator so must be imported later

# This is a type defined in internals, so can include it here
# TODO: Move this around to be somewhere sensible
from ..internals import Val_
make_VT("Val", pytype=Val_)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import make_VT
from .helpers import remove_names, absorbed

def rp_validation(self):
    abs = remove_names(absorbed(self))
    if not len(abs)!=3:
        raise TypeError(f"`RP`[...]  must be initialized with a triple to match on. Got x={abs}")
    return True

def operates_on_ctor(x):
    from .._ops import operates_on, contains
    from . import Is
    return Is[operates_on | contains[x]]

def related_ops_ctor(x):
    from . import Is
    from .._ops import related_ops, contains
    return Is[related_ops | contains[x]]

def used_for_ctor(x):
    from . import Is
    from .._ops import used_for, contains
    return Is[used_for | contains[x]]

# TODO: is_a for everything above


RP          = make_VT('RP')
HasValue    = make_VT('HasValue')   # TODO Remove this later
Query      = make_VT('Query'    )
OperatesOn  = make_VT('OperatesOn',   constructor_func = operates_on_ctor)#,  get_item_func = operates_on_ctor)
RelatedOps  = make_VT('RelatedOps',   constructor_func = related_ops_ctor)#,  get_item_func = related_ops_ctor)
UsedFor     = make_VT('UsedFor',      constructor_func = used_for_ctor)#, get_item_func =used_for_ctor)






# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import make_VT

from .helpers import remove_names, absorbed

def same_as_validation(typ):
    abs = remove_names(absorbed(typ))
    if len(abs) != 1:
        raise Exception(f"SameAs needs exactly one absorbed argument, have {abs}")
    return True

def same_as_is_a(x, typ):
    assert same_as_validation(typ)
    
    thing = remove_names(absorbed(typ))[0]
    return discard_frame_imp(x) == discard_frame_imp(thing)


SameAs      = make_VT('SameAs', is_a_func = same_as_is_a)
# TODO Add or remove these once decided their importance
# Tagged     = make_VT('Tagged'   )
# LazyValue  = make_VT('LazyValue')
# Function   = make_VT('Function' )
# GraphDelta = make_VT('GraphDelta')



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import make_VT

# Show of these are not python types but are very close and fit into the category of scalar primitives

make_VT('Nil', pytype=type(None))
make_VT('Any',
              is_a_func=lambda x,typ: True,
              is_subtype_func=lambda x,typ: True)

def numeric_is_a(x, typ):
    from .value_type import _value_type_pytypes
    python_type = _value_type_pytypes[typ._d['type_name']]
    try:
        return isinstance(x, python_type) or python_type(x) == x
    except:
        return False

PyBool = make_VT("PyBool", pytype=bool)
Bool = make_VT("Bool", pytype=bool, is_a_func=numeric_is_a)

PyInt = make_VT("PyInt", pytype=int)
Int = make_VT("Int", pytype=int, is_a_func=numeric_is_a)

PyFloat = make_VT("PyFloat", pytype=float)
Float = make_VT("Float", pytype=float, is_a_func=numeric_is_a)

String = make_VT('String', pytype=str)
PyBytes = make_VT('PyBytes', pytype=bytes)

from frozendict import frozendict
PyList = make_VT('PyList', pytype=list)
def PyDict_is_a(x, typ):
    return isinstance(x, (dict, frozendict))
PyDict = make_VT('PyDict', pytype=dict, is_a_func=PyDict_is_a)
PySet = make_VT('PySet', pytype=set)

PyTuple = make_VT('PyTuple', pytype=tuple)




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *


class MyTestCase(unittest.TestCase):


    def test_dict_syntax(self):
        fg = FlatGraph([{
                ET.Person['z1'] : {
                RT.FirstName: "Fred",
                RT.YearOfBirth: 1970,
                RT.Parent: {
                    ET.Man: {
                        RT.Name: "Bob"
                    }
                }
            }
        }])

        g2 = Graph()
        fg | transact[g2] | run


    def test_mixed_types(self):
        fg = FlatGraph()
        fg = (fg 
        | insert[[
            ET.Foo['x1'],
            AET.Int | assign[41],
            AET.Int['k2'] | assign[41],
            (Any['x1'], RT.A['r2'], ET.Baz),
            (Any['r2'], RT.B, AET.String| assign["Fred"]),
            (Any['r2'], RT.C, AET.String['aet1'] | assign["Some"]),
            (Any['r2'], RT.G, "Egypt"),
            {ET.Person['p1']:{                             
            RT.Name: "Die Antwoord",
            RT.Style: "Rap-Rave",
            RT.Member['r1']: 5,}},
            (Any['p1'], RT.Name, "Ninja"),
        ]]
        | collect)


        g2 = Graph()
        fg | transact[g2] | run


    def test_all_types(self):
        g  = Graph()
        z0 = ET.Person | g | run
        z1 = EntityRef(z0)
        z2 = AttributeEntityRef(AET.Bool | g | run)
        rt = ((z0, RT.Name, AET.String) | g | run)[1]
        z3 = RelationRef(((rt, RT.Another, AET.String) | g | run)[1])
        z4 = AET.String | g | run
        z5 = (z4 | assign["hey"]) | g | run 

        fg = FlatGraph()
        fg = (fg 
        | insert[ET.Foo['x1']] 
        | insert[(AET.Int | assign[41])] 
        | insert[AET.Int['k2'] | assign[42] ] 
        | insert[(Any['x1'], RT.A['r2'], ET.Baz)]
        | insert[(Any['r2'], RT.B, AET.String| assign["Fred"])]    # assign value in one go
        | insert[(Any['r2'], RT.C, AET.String['aet1'] | assign["Some"])]    # internal id
        | insert[(Any['r2'], RT.G, "Egypt")]                  # value node
        | insert[AET.String] 
        | insert[AET.String['n1'] | assign["zeyad"]]
        | insert[AET.String | assign["wow"]]
        | insert[AET.Int['n2'] | assign[45]]
        | insert[Any['n2'] | assign[8]]
        | insert[ET.Person['p1']] 
        | insert[Val("r2")]
        | insert[Val(ET.Foo)]
        | insert[Val("English")]
        | insert[(ET.Cat['c1'], RT.D['r1'], Any['n1'])]
        | insert[(Any['r1'], RT.E, Val("English"))]
        | insert[(Any['r1'], RT.F, Any['r2'])]
        | insert[z4 | assign['500000'] ]
        | insert[z5]
        | insert[z1] 
        | insert[(z1, RT.H['t1'], Any['n1'])]
        | insert[RelationRef(rt)]
        | insert[z3]
        | insert[z2]
        | insert[z2 | assign[True]]
        | collect
        )

        # Transact Test
        g2 = Graph()
        fg | transact[g2] | run


    def test_fg_serialization_and_deserialization(self):
        fg = FlatGraph([{
                ET.Person['z1'] : {
                RT.FirstName: "Fred",
                RT.YearOfBirth: 1970,
                RT.Parent: {
                    ET.Man: {
                        RT.Name: "Bob"
                    }
                }
            }
        }])
        fg | to_json | from_json | collect 


    def test_fg_graphviz(self):
        fg = FlatGraph([{
                ET.Person['z1'] : {
                RT.FirstName: "Fred",
                RT.YearOfBirth: 1970,
                RT.Parent: {
                    ET.Man: {
                        RT.Name: "Bob"
                    }
                }
            }
        }])
        fg | graphviz | collect


    def test_merging_flatgraphs(self):
        g  = Graph()
        z0 = ET.Person | g | run
        
        fg = FlatGraph([{
                ET.Person['z1'] : {
                RT.FirstName: "Fred",
                RT.YearOfBirth: 1970,
                RT.Friend: z0,
                RT.Parent: {
                    ET.Man: {
                        RT.Name: "Bob"
                    }
                }
            }
        }])

        fg2 = FlatGraph()
        fg2 = (fg2 
            | insert[(ET.Foo['x1'], RT.X, z0)]       
            | insert[(ET.Cat, RT.Y['r1'],AET.String['n1'] |assign["42"])]    
            | insert[(Any['r1'], RT.Z, Val("English"))]   
            | insert[Val("Zeyad")]
            | insert[AET.Float['f1']] 
            | collect
        )

        merge(fg, fg2)


if __name__ == '__main__':
    unittest.main()


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: This should never depend on core at the top level!

from ... import report_import
report_import("zef.core.internals")

from ...pyzef.main import (
    EntityType,
    RelationType,
    Graph,
    GraphRef,
    ZefRef,
    EZefRef,
    ZefEnumValue,
)

from ...pyzef.internals import (
    AttributeEntityType,
    AttributeEntityTypeStruct_QuantityFloat,
    AttributeEntityTypeStruct_QuantityInt,
    AttributeEntityTypeStruct_Enum,
    BaseUID,
    BlobType,
    BlobTypeStruct,
    Delegate,
    DelegateRelationTriple,
    EntityTypeStruct,
    EternalUID,
    FinishTransaction,
    GraphData,
    Graph_from_ptr,
    KW,
    MMAP_STYLE_ANONYMOUS,
    MMAP_STYLE_AUTO,
    MMAP_STYLE_FILE_BACKED,
    NamedHeadRange,
    ObservablesDictElement,
    RelationTypeStruct,
    StartTransactionReturnTx,
    Subscription,
    UpdateHeads,
    UpdatePayload,
    ValueRepType,
    ValueRepTypeStruct_QuantityFloat,
    ValueRepTypeStruct_QuantityInt,
    ValueRepTypeStruct_Enum,
    VRT,
    ZefEnumStruct,
    ZefEnumStructPartial,
    ZefRefUID,
    add_entity_type,
    add_enum_type,
    add_relation_type,
    add_keyword,
    all_entity_types,
    all_enum_types_and_values,
    all_relation_types,
    apply_update,
    blob_to_json,
    compress_zstd,
    create_graph_from_bytes,
    create_update_heads,
    create_update_payload,
    created_token_list,
    current_zefdb_protocol_version,
    decompress_zstd,
    delete_graphdata,
    delegate_to_ezr,
    early_token_list,
    ezr_to_delegate,
    get_data_layout_version_info,
    get_enum_value_from_string,
    get_graph_revision_info,
    get_latest_complete_tx_node,
    get_local_process_graph,
    graph_as_UpdatePayload,
    gtd_info_str,
    has_delegate,
    has_uid,
    heads_apply,
    initialise_butler,
    initialise_butler_as_master,
    is_BaseUID,
    is_EternalUID,
    is_ZefRefUID,
    is_vrt_a_enum,
    is_vrt_a_quantity_float,
    is_vrt_a_quantity_int,
    is_any_UID,
    is_delegate,
    is_delegate_relation_group,
    is_root,
    is_up_to_date,
    list_graph_manager_uids,
    login,
    logout,
    make_random_uid,
    max_zefdb_protocol_version,
    #memory_details,
    merge_atomic_entity_,
    merge_entity_,
    merge_relation_,
    num_blob_indexes_to_move,
    pageout,
    parse_payload_update_heads,
    partial_hash,
    root_node_blob_index,
    search_value_node,
    set_data_layout_version_info,
    set_graph_revision_info,
    show_blob_details,
    size_of_blob,
    start_connection,
    stop_butler,
    stop_connection,
    to_uid,
    validate_message_version,
    value_hash,
    wait_for_auth,
)

from ...pyzef.verification import (
    verify_graph
)
from ...pyzef.zefops import (
    SerializedValue
)

from ...pyzef.main import ZefRef, zwitch

# ----------------------------- monkey patch to allow dispatching on enums ---------------------------------------
# For some reason we need to pass it through a lambda fct to get the bindings to work.
# BlobType.__rrshift__ = lambda bt, any_zr_type: fct__rrshift__(bt, any_zr_type)
# BlobType.__rlshift__ = lambda bt, any_zr_type: fct__rlshift__(bt, any_zr_type)
# BlobType.__gt__ = lambda bt, any_zr_type: fct__gt__(bt, any_zr_type)
# BlobType.__lt__ = lambda bt, any_zr_type: fct__lt__(bt, any_zr_type)

BlobType.__str__ = lambda self: repr(self).split(':')[0][10:]   # strip away the 'BlobType. ...' and out <>

from .rel_ent_classes import ET, RT, EN, AET
from .merges import register_merge_handler
register_merge_handler()
from .schema import register_schema_validator
register_schema_validator()
from .value_type_check import register_value_type_check
register_value_type_check()
from .determine_primitive_type import register_determine_primitive_type
register_determine_primitive_type()

BT = BlobTypeStruct()

def safe_current_task():
    import asyncio
    try:
        return asyncio.current_task()
    except RuntimeError:
        return None
    
global_transaction_task = {}
from contextlib import contextmanager
@contextmanager
def Transaction(g, wait=None, rollback_empty=None, check_schema=None):
    global global_transaction_task
    from ...pyzef.zefops import uid

    if wait is None:
        wait = zwitch.default_wait_for_tx_finish()
    if rollback_empty is None:
        rollback_empty = zwitch.default_rollback_empty_tx()
    if check_schema is None:
        check_schema = True

    cur_task = safe_current_task()

    if cur_task is None:
        current_tx = StartTransactionReturnTx(g)
        try:
            yield current_tx
        finally:
            FinishTransaction(g, wait, rollback_empty, check_schema)
    else:
        prev_val = global_transaction_task.get(uid(g), None)

        if prev_val is not None and prev_val != cur_task:
            raise Exception("Can't open a Transaction from a different task to the original Transaction. Note that ZefDB is not safe to create simultaneous transactions from different asyncio tasks. Ideally, there should be no asyncio task switch occuring during a transaction.")

        global_transaction_task[uid(g)] = cur_task
        current_tx = StartTransactionReturnTx(g)
        try:
            yield current_tx
        finally:
            if prev_val is None and uid(g) in global_transaction_task:
                del global_transaction_task[uid(g)]
            else:
                global_transaction_task[uid(g)] = prev_val

            FinishTransaction(g, wait, rollback_empty, check_schema)



def assign_value_imp(z, value):
    from .._ops import is_a
    from ...pyzef.zefops import SerializedValue, assign_value as c_assign_value
    from ..VT import ValueType_, AET
    from .. import VT
    from .rel_ent_classes import AET as internal_AET
    from ..graph_delta import scalar_types

    assert isinstance(z, ZefRef) or isinstance(z, EZefRef)

    # We can't be sure what kind of zefref we have, and if it is complex things
    # break at the moment, so do this thoroughly here.
    if not is_a(z, VT.AET[VT.Any]):
        print(z)
        raise Exception("E/ZefRef is not an AET!")
    aet = VT.AET(z)
    if isinstance(value, VT.AET):
        value = get_c_token(value)
    if isinstance(value, ValueType_):
        value = internal_AET[value]

    if not isinstance(value, scalar_types):
        value = SerializedValue.serialize(value)
    elif get_c_token(aet).rep_type == VRT.Serialized:
        value = SerializedValue.serialize(value)
    try:
        c_assign_value(z, value)
    except Exception as exc:
        print(f"There was an exception in the c call for assign value. z={z} and value={value} and aet={aet}")
        raise


def instantiate_value_node_imp(value, g):
    from ...pyzef.zefops import SerializedValue
    from ...pyzef.main import instantiate_value_node as c_instantiate_value_node
    from .._ops import is_a
    from ..VT import ValueType_
    from .. import VT
    from ..graph_delta import scalar_types

    from .rel_ent_classes import AET as internal_AET
    if isinstance(value, VT.AET):
        value = get_c_token(value)
    elif isinstance(value, ValueType_):
        value = internal_AET[value]

    if not isinstance(value, scalar_types):
        value = SerializedValue.serialize(value)
    return c_instantiate_value_node(value, g)


def get_c_token(x):
    from ..VT.rae_types import RAET_get_token
    token = RAET_get_token(x)
    from ..VT import ValueType
    if isinstance(token, ValueType):
        from .rel_ent_classes import AET as internal_AET
        return internal_AET[token]
    return token


from dataclasses import dataclass
@dataclass(frozen=True)
class Val_:
    # arg: VT.Any
    arg: int
    iid: str = None

    def __getitem__(self, x):
        if self.iid is not None:
            raise Exception("Can't overwrite iid")
        return Val_(self.arg, x)

    def __hash__(self):
        from ..VT.value_type import hash_frozen
        return hash_frozen(("Val_", self.arg, self.iid))


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import make_VT, insert_VT, ValueType, SetOf, BlobPtr
from .value_type import is_empty_, is_type_name_
from .helpers import absorbed, type_name
from .. import internals

# These pop up a lot as special cases
AET_QFloat = make_VT('AET_QFloat', pytype=internals.AttributeEntityTypeStruct_QuantityFloat)
AET_QInt = make_VT('AET_QInt', pytype=internals.AttributeEntityTypeStruct_QuantityInt)
AET_Enum = make_VT('AET_Enum', pytype=internals.AttributeEntityTypeStruct_Enum)

VRT_QFloat = make_VT('VRT_QFloat', pytype=internals.ValueRepTypeStruct_QuantityFloat)
VRT_QInt = make_VT('VRT_QInt', pytype=internals.ValueRepTypeStruct_QuantityInt)
VRT_Enum = make_VT('VRT_Enum', pytype=internals.ValueRepTypeStruct_Enum)

# Helpers

def RAET_get_token(typ):
    # This allows anything in the absorbed and takes just the first likely item
    # This is far too lenient, it will likely break in the future
    opts = [x for x in absorbed(typ) if isinstance(x, (EntityTypeToken, RelationTypeToken, BlobTypeToken, AttributeEntityTypeToken, ValueRepTypeToken,
                                                       AET_QFloat, AET_QInt, AET_Enum,
                                                       ValueType))]
    if len(opts) == 0:
        return None
    return opts[0]

def RAET_get_names(typ):
    token = RAET_get_token(typ)
    if token is not None:
        names = list(absorbed(typ))
        names.remove(token)
        return tuple(names)
    else:
        return absorbed(typ)

def wrap_attr_readonly_token(orig):
    def this_get_attr(self, name):
        token = RAET_get_token(self)
        if token is None:
            out = getattr(orig, name)
        else:
            # This is just for AETs
            if isinstance(token, AET_QFloat):
                out = getattr(orig.QuantityFloat, name)
            elif isinstance(token, AET_QInt):
                out = getattr(orig.QuantityInt, name)
            elif isinstance(token, AET_Enum):
                out = getattr(orig.Enum, name)
            else:
                raise AttributeError(name)
        assert len(RAET_get_names(self)) == 0, "Going to lose names"
        base_token_type = self._replace(absorbed=())
        return base_token_type[out]
    def this_dir(self):
        token = RAET_get_token(self)
        if token is not None:
            # This is just for AETs
            if isinstance(token, AET_QFloat):
                return dir(orig.QuantityFloat)
            elif isinstance(token, AET_QInt):
                return dir(orig.QuantityInt)
            elif isinstance(token, AET_Enum):
                return dir(orig.Enum)
            else:
                return []
        return dir(orig)
    return (this_get_attr, None, this_dir)

def token_subtype(other, this):
    if type_name(other) != type_name(this):
        return False
    other_token = RAET_get_token(other)
    this_token = RAET_get_token(this)
    if other_token is None:
        return this_token is None
    if this_token is None:
        return True
    if isinstance(this_token, ValueType):
        if isinstance(other_token, ValueType):
            return issubclass(other_token, this_token)
        return isinstance(other_token, this_token)
    return other_token == this_token

def token_validation(self, token_type):
    # my_name = type_name(self)

    # abs = remove_names(absorbed(self))
    # if len(abs) == 0:
    #     return True
    # if len(abs) >= 2:
    #     raise Exception(f"Should only have one token absorbed into a {my_name}")
    # thing = abs[0]

    # if not isinstance(thing, token_type):
    #     raise Exception(f"A {my_name} doesn't contain a {token_type} but has a {thing} instead")

    # Anything is allowed now
    return True

def token_str(self):
    my_name = self._d["type_name"]
    s = my_name
    token = RAET_get_token(self)
    if token is not None:
        if isinstance(token, (EntityTypeToken, RelationTypeToken, AttributeEntityTypeToken)):
            s += "." + token.name
        elif isinstance(token, BlobTypeToken):
            s += "." + str(token)
        else:
            s += "[" + str(token) + "]"

    for thing in RAET_get_names(self):
        s += f"[{thing!r}]"
    return s

def ET_is_a(x, typ):
    from . import DelegateRef, EntityRef
    token = RAET_get_token(typ)
    if token is None:
        if isinstance(x, DelegateRef):
            print(" ⚠️  Warning, `ET` was used as the type in an isinstance comparison for an `Entity`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., Entity) instead. If you want to only catch e.g. isinstance(ET.X, ET) then until the deprecation occurs, you should use isinstance(ET.X, ValueType & ET). ⚠ ️")
            return isinstance(x.item, EntityTypeToken)
        elif isinstance(x, BlobPtr):
            print(" ⚠ ️ Warning, `ET` was used as the type in an isinstance comparison for an `Entity`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., Entity) instead. If you want to only catch e.g. isinstance(ET.X, ET) then until the deprecation occurs, you should use isinstance(ET.X, ValueType & ET). ⚠ ️")
            return internals.BT(x) == internals.BT.ENTITY_NODE
        elif isinstance(x, EntityRef):
            print(" ⚠ ️ Warning, `ET` was used as the type in an isinstance comparison for an `Entity`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., Entity) instead. If you want to only catch e.g. isinstance(ET.X, ET) then until the deprecation occurs, you should use isinstance(ET.X, ValueType & ET). ⚠ ️")
            return True
        return isinstance(x, ValueType) and type_name(x) == "ET"
    else:
        if isinstance(x, DelegateRef):
            return x.item == token
        elif isinstance(x, BlobPtr):
            if internals.BT(x) != internals.BT.ENTITY_NODE:
                return False
            if isinstance(token, ValueType):
                return isinstance(internals.ET(x), token)
            return internals.ET(x) == token
        elif isinstance(x, EntityRef):
            from .._ops import rae_type
            return RAET_get_token(rae_type(x)) == token
        else:
            return False

def AET_is_a(x, typ):
    from . import DelegateRef, AttributeEntityRef
    token = RAET_get_token(typ)
    if token is None:
        if isinstance(x, DelegateRef):
            print(" ⚠ ️ Warning, `AET` was used as the type in an isinstance comparison for an `AttributeEntity`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., AttributeEntity) instead. If you want to only catch e.g. isinstance(AET.Int, AET) then until the deprecation occurs, you should use isinstance(AET.Int, ValueType & AET). ⚠ ️")
            return isinstance(x.item, AttributeEntityTypeToken)
        elif isinstance(x, BlobPtr):
            print(" ⚠ ️ Warning, `AET` was used as the type in an isinstance comparison for an `AttributeEntity`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., AttributeEntity) instead. If you want to only catch e.g. isinstance(AET.Int, AET) then until the deprecation occurs, you should use isinstance(AET.Int, ValueType & AET). ⚠ ️")
            return internals.BT(x) == internals.BT.ATTRIBUTE_ENTITY_NODE
        elif isinstance(x, AttributeEntityRef):
            print(" ⚠ ️ Warning, `AET` was used as the type in an isinstance comparison for an `AttributeEntity`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., AttributeEntity) instead. If you want to only catch e.g. isinstance(AET.Int, AET) then until the deprecation occurs, you should use isinstance(AET.Int, ValueType & AET). ⚠ ️")
            return True
        return is_type_name_(x, "AET")
    else:
        if isinstance(x, DelegateRef):
            x_aet = x.item
        elif isinstance(x, BlobPtr):
            if internals.BT(x) != internals.BT.ATTRIBUTE_ENTITY_NODE:
                return False
            x_aet = internals.AET(x)
        elif isinstance(x, ValueType) and type_name(x) == "AET":
            x_aet = RAET_get_token(x)
            if x_aet is None:
                return False
        elif isinstance(x, (AttributeEntityTypeToken, AET_QFloat, AET_QInt, AET_Enum)):
            x_aet = x
        elif isinstance(x, AttributeEntityRef):
            from .._ops import rae_type
            x_aet = rae_type(x)
        else:
            return False

        if isinstance(token, ValueType):
            return isinstance(x_aet, token)

        if token == x_aet:
            return True

        if isinstance(token, AttributeEntityTypeToken) and token.complex_value is not None:
            raise Exception(f"Checking isinstance on complex AETs (got {typ}) is not yet implemented. Coming soon!")
        if isinstance(x_aet, AttributeEntityTypeToken) and x_aet.complex_value is not None:
            raise Exception(f"Checking isinstance on complex AETs (got {x_aet}) is not yet implemented. Coming soon!")

        if isinstance(token, (AET_QFloat, AET_QInt, AET_Enum)):
            if isinstance(x_aet, AttributeEntityTypeToken):
                if isinstance(token, AET_QFloat):
                    return internals.is_vrt_a_quantity_float(x_aet.rep_type)
                if isinstance(token, AET_QInt):
                    return internals.is_vrt_a_quantity_int(x_aet.rep_type)
                if isinstance(token, AET_QEnum):
                    return internals.is_vrt_a_enum(x_aet.rep_type)
        return False
def RT_is_a(x, typ):
    from . import DelegateRef, RelationRef
    token = RAET_get_token(typ)
    if token is None:
        if isinstance(x, DelegateRef):
            print(" ⚠ ️ Warning, `RT` was used as the type in an isinstance comparison for a `Relation`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., Relation) instead. If you want to only catch e.g. isinstance(RT.X, RT) then until the deprecation occurs, you should use isinstance(RT.X, ValueType & RT). ⚠ ️")
            if type(x.item) == internals.DelegateRelationTriple:
                return isinstance(x.item.rt, RelationTypeToken)
            return isinstance(x.item, RelationTypeToken)
        elif isinstance(x, BlobPtr):
            print(" ⚠ ️ Warning, `RT` was used as the type in an isinstance comparison for a `Relation`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., Relation) instead. If you want to only catch e.g. isinstance(RT.X, RT) then until the deprecation occurs, you should use isinstance(RT.X, ValueType & RT). ⚠ ️")
            return internals.BT(x) == internals.BT.RELATION_EDGE
        elif isinstance(x, RelationRef):
            print(" ⚠ ️ Warning, `RT` was used as the type in an isinstance comparison for a `Relation`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., Relation) instead. If you want to only catch e.g. isinstance(RT.X, RT) then until the deprecation occurs, you should use isinstance(RT.X, ValueType & RT). ⚠ ️")
            return True
        return isinstance(x, ValueType) and type_name(x) == "RT"
    else:
        if isinstance(x, DelegateRef):
            if type(x.item) == internals.DelegateRelationTriple:
                return x.item.rt == token
            else:
                return x.item == token
        elif isinstance(x, BlobPtr):
            if internals.BT(x) != internals.BT.RELATION_EDGE:
                return False
            # TODO: EntityTypeToken
            if isinstance(token, ValueType):
                return isinstance(internals.RT(x), token)
            return internals.RT(x) == token
        elif isinstance(x, RelationRef):
            from .._ops import rae_type
            return RAET_get_token(rae_type(x)) == token
        else:
            return False
def BT_is_a(x, typ):
    token = RAET_get_token(typ)
    if token is None:
        if isinstance(x, BlobPtr):
            print(" ⚠ ️ Warning, `BT` was used as the type in an isinstance comparison for a `BlobPtr`. This will become False in the future, but currently returns True. Update your code to compare isinstance(..., BlobPtr) instead. If you want to only catch e.g. isinstance(BT.VALUE_NODE, BT) then until the deprecation occurs, you should use isinstance(BT.VALUE_NODE, ValueType & BT). ⚠ ️️")
            return True
        return isinstance(x, ValueType) and type_name(x) == "BT"
    else:
        c_bt = token
        if not isinstance(c_bt, BlobTypeToken):
            raise Exception("TODO")
        if isinstance(x, BlobPtr):
            return c_bt == internals.BT(x)
        if c_bt == internals.BT.RELATION_EDGE:
            if isinstance(x, RT):
                return True
        if c_bt == internals.BT.ENTITY_NODE:
            if isinstance(x, ET):
                return True
        if c_bt == internals.BT.ATTRIBUTE_ENTITY_NODE:
            if isinstance(x, AET):
                return True
        return False
def VRT_is_a(x, typ):
    from . import DelegateRef
    token = RAET_get_token(typ)
    if token is None:
        if isinstance(x, DelegateRef):
            print(" ⚠ ️ Warning, `VRT` was used as the type in an isinstance comparison for a `Delegate`. This will become False in the future, but currently returns True. This actually is nonsensical, update your code logic. ⚠ ️")
            return isinstance(x.item, ValueRepTypeToken)
        elif isinstance(x, BlobPtr):
            print(" ⚠ ️ Warning, `VRT` was used as the type in an isinstance comparison for a `Delegate`. This will become False in the future, but currently returns True. This actually is nonsensical, update your code logic. ⚠ ️")
            return internals.BT(x) in [internals.BT.ATTRIBUTE_ENTITY_NODE,
                                       internals.BT.VALUE_NODE]
        return is_type_name_(x, "VRT")
    else:
        if isinstance(x, DelegateRef):
            x_vrt = x.item
        elif isinstance(x, BlobPtr):
            if internals.BT(x) not in [internals.BT.ATTRIBUTE_ENTITY_NODE,
                                       internals.BT.VALUE_NODE]:
                return False
            x_vrt = internals.VRT(x)
        elif isinstance(x, ValueType) and type_name(x) == "VRT":
            x_vrt = RAET_get_token(x)
            if x_vrt is None:
                return False
        elif isinstance(x, (ValueRepTypeToken, VRT_QFloat, VRT_QInt, VRT_Enum)):
            x_vrt = x
        else:
            return False

        if isinstance(token, ValueType):
            return isinstance(x_vrt, token)

        if token == x_vrt:
            return True

        if isinstance(token, (VRT_QFloat, VRT_QInt, VRT_Enum)):
            if isinstance(x_aet, AttributeEntityTypeToken):
                if isinstance(token, VRT_QFloat):
                    return internals.is_vrt_a_quantity_float(x_vrt)
                if isinstance(token, VRT_QInt):
                    return internals.is_vrt_a_quantity_int(x_vrt)
                if isinstance(token, VRT_QEnum):
                    return internals.is_vrt_a_enum(x_vrt)
        return False


def ET_ctor(self, *args, **kwargs):
    if RAET_get_token(self) is None:
        return ET[internals.ET(*args, **kwargs)]
    else:
        return EntityValueInstance(self, *args, **kwargs)
    
# TODO: Move this somewhere
from ..patching import EntityValueInstance_
EntityValueInstance = make_VT('EntityValueInstance', pytype=EntityValueInstance_)
EntityTypeToken = make_VT('EntityTypeToken', pytype=internals.EntityType)

ET = make_VT('ET',
             constructor_func=ET_ctor,
             pass_self=True,
             attr_funcs=wrap_attr_readonly_token(internals.ET),
             is_a_func=ET_is_a,
             is_subtype_func=token_subtype,
             str_func=token_str)

AttributeEntityTypeToken = make_VT('AttributeEntityTypeToken', pytype=internals.AttributeEntityType)
def AET_ctor(self, x):
    if type(x) == str:
        return getattr(self, x)
    token = RAET_get_token(self)
    if token is not None:
        return NotImplemented
    return AET[internals.AET(x)]

AET = make_VT('AET',
              constructor_func=AET_ctor,
              pass_self=True,
              attr_funcs=wrap_attr_readonly_token(internals.AET),
              is_a_func=AET_is_a,
              is_subtype_func=token_subtype,
              str_func=token_str)

RelationTypeToken = make_VT('RelationTypeToken', pytype=internals.RelationType)
RT = make_VT('RT',
             constructor_func=lambda x: RT[internals.RT(x)],
             attr_funcs=wrap_attr_readonly_token(internals.RT),
             is_a_func=RT_is_a,
             is_subtype_func=token_subtype,
             str_func=token_str)


BlobTypeToken = make_VT('BlobTypeToken', pytype=internals.BlobType)
BT = make_VT('BT',
             constructor_func=lambda x: BT[internals.BT(x)],
             attr_funcs=wrap_attr_readonly_token(internals.BT),
             is_a_func=BT_is_a,
             is_subtype_func=token_subtype,
             str_func=token_str)

# BT         = ValueType_(type_name='BT',   constructor_func=pyzef.internals.BT, attr_funcs=wrap_attr_readonly(internals.BT, None), pytype=internals.BlobType)

ValueRepTypeToken = make_VT('ValueRepTypeToken', pytype=internals.ValueRepType)
def VRT_ctor(self, x):
    if type(x) == str:
        return getattr(self, x)
    token = RAET_get_token(self)
    if token is not None:
        return NotImplemented
    return VRT[internals.VRT(x)]
VRT = make_VT('VRT',
              constructor_func=VRT_ctor,
              pass_self=True,
              attr_funcs=wrap_attr_readonly_token(internals.VRT),
              is_a_func=VRT_is_a,
              is_subtype_func=token_subtype,
              str_func=token_str)


def tx_is_a(x, typ):
    return isinstance(x, BT.TX_EVENT_NODE)

TX = make_VT("TX", is_a_func=tx_is_a)

RAET = insert_VT("RAET", ET | RT | AET)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import _version
__version__ = _version.get_versions()['version']

########################################################
# * Exposing common functions
#------------------------------------------------------

import_order = []
def report_import(x):
    if x in import_order:
        return
    import_order.append(x)
report_import("zef")

# This set of imports is to define the order. Later imports are the ones to
# actually provide useful exports.
from . import core
from . import pyzef
from . import ops
from . import ui

from .core import *

pyzef.internals.finished_loading_python_core()

############################################
# * Starting the butler
#------------------------------------------

# Putting this into a function to avoid polluting namespace
def _autostart_behaviour():
    auto_start = ops.config("butler.autoStart", KW.get)
    if auto_start:
        core.internals.initialise_butler()

_autostart_behaviour()

# We always run this, in case the user has started the butler manually instead of automatically
import atexit
@atexit.register
def _stop_butler():
    from . import pyzef
    pyzef.internals.stop_butler()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



# Functions that are useful for implementing other type functions

# Note: must be imported after python_types
from .python_types import PyTuple
from . import ValueType
from .value_type import absorbed, type_name, is_subtype_

def remove_names(absorbed: PyTuple):
    from . import Variable
    # TODO: Replace with early ops if possible
    return tuple(x for x in absorbed if not isinstance(x, Variable))

def names_of(typ: ValueType):
    from . import Variable
    # TODO: Replace with early ops if possible
    return tuple(x for x in absorbed(typ) if isinstance(x, Variable))


##############################
# * Subtype VTs

# These VTs have a single curried type which is used for representing them and
# for membership tests.
#----------------------------

# def generic_subtype_str(self):
#     s = self._d["type_name"]
#     if "subtype" in self._d:
#         s += f"[{self._d['subtype']}]"
#     s += ''.join(f"[{x!r}]" for x in self._d["absorbed"])
#     return s

def generic_subtype_validate(typ):
    # Should validation return True/False? Or True/String? Or True and throw?
    items = remove_names(absorbed(typ))
    if len(items) == 0:
        return True
    if len(items) > 2:
        raise Exception(f"Generic subtype validation failed: there are two or more absorbed subtypes ({items})")
    subtype = items[0]
    if not isinstance(subtype, ValueType):
        raise Exception(f"Generic subtype validation failed: subtype is not a ValueType ({subtype})")
    return True

def generic_subtype_get(typ):
    assert generic_subtype_validate(typ)
    items = remove_names(absorbed(typ))
    if len(items) == 0:
        return None
    else:
        return items[0]


################################################
# * Covariant subtype VTs

# These subtype VTs allow for multiple parameters and the membership test is
# performed covariantly (that is, the individual parameters are compared and
# combined with AND)
#----------------------------------------------

def generic_covariant_validate(typ):
    # Should validation return True/False? Or True/String? Or True and throw?
    items = remove_names(absorbed(typ))
    for item in items:
        if not isinstance(item, ValueType):
            raise Exception(f"Generic covariant type validation failed: item is not a ValueType ({item})")
    return True

def generic_covariant_parameters(typ):
    assert generic_covariant_validate(typ)
    return remove_names(absorbed(typ))

def generic_covariant_is_subtype(x, super):
    if type_name(x) != type_name(super):
        return False
    super_params = generic_covariant_parameters(super)
    x_params = generic_covariant_parameters(x)

    return all(is_subtype_(a,b) is True for a,b in zip(x_params, super_params))


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .VT import *
from .VT import make_VT
from . import internals
from ..pyzef import zefops as pyzefops

def to_delegate_imp(first_arg, *curried_args):
    # TODO: Break this up and document

    if isinstance(first_arg, DelegateRef):
        if len(curried_args) == 0:
            return first_arg
        if isinstance(curried_args[0], Graph):
            g = curried_args[0]
        else:
            assert isinstance(curried_args[0], GraphSlice)
            g = Graph(curried_args[0])
        if len(curried_args) == 1:
            d_ezr = internals.delegate_to_ezr(first_arg, g, False, 0)
        elif len(curried_args) == 2:
            d_ezr = internals.delegate_to_ezr(first_arg, g, curried_args[1], 0)
        else:
            raise Exception("Too many args for to_delegate with a Delegate")

        if d_ezr is None:
            return None
        elif isinstance(curried_args[0], Graph):
            return d_ezr
        else:
            from ._ops import in_frame, exists_at
            if not exists_at(d_ezr, curried_args[0]):
                raise Exception("Delegate does not exist at given graph slice.")
            return in_frame(d_ezr, curried_args[0])

    if isinstance(first_arg, ZefRef) or isinstance(first_arg, EZefRef):
        assert len(curried_args) == 0
        return internals.ezr_to_delegate(first_arg)

    raise Exception(f"Unknown type for to_delegate: {type(first_arg)}. Maybe you meant to write delegate_of?")

# This is for internal use only - it tries to convert a tuple or single RAET
# into a delegate of order zero.
def attempt_to_delegate(args):
    if isinstance(args, tuple):
        assert len(args) == 3
        args = tuple(internals.get_c_token(x) if isinstance(x, ValueType) else x for x in args)
        return DelegateRef(DelegateRef(args[0]), args[1], DelegateRef(args[2]))
    else:
        args = internals.get_c_token(args) if isinstance(args, ValueType) else args
        return DelegateRef(args)

def delegate_of_imp(x, arg1=None, arg2=None):
    # TODO: Move implementation
    from ._ops import in_frame, frame, collect

    # TODO: Break this up and document
    if isinstance(x, EZefRef) or isinstance(x, ZefRef):
        assert arg2 is None
        if arg1 is None:
            create = False
        else:
            create = arg1
        assert isinstance(create, Bool)

        d = pyzefops.delegate_of(to_delegate_imp(x))
        z = to_delegate_imp(d, Graph(x), create)
        if z is None:
            return None
        if isinstance(x, ZefRef):
            z = z | in_frame[frame(x)] | collect
        return z

    if isinstance(x, DelegateRef):
        if arg1 is None:
            create = False
            g = None
        else:
            g = arg1
            if arg2 is None:
                create = False
            else:
                create = arg2

        d = pyzefops.delegate_of(x)
        if g is None:
            return d
        else:
            return to_delegate_imp(d, g, create)

    # Fallback
    return delegate_of_imp(attempt_to_delegate(x), arg1, arg2)


def DelegateRef_ctor(*args):
    from .VT.helpers import type_name
    from .VT.rae_types import RAET_get_token
    if len(args) < 1 or len(args) > 3:
        raise Exception("DelegateRef constructor requires 1, 2 or 3 arguments")
    def to_c_token_maybe(x):
        if isinstance(x, ValueType) and type_name(x) in ["ET", "RT", "VRT"]:
            return RAET_get_token(x)
        return x

    args = [to_c_token_maybe(x) for x in args]
    return internals.Delegate(*args)

# TODO: Need a is_a for this too - or rather make this the main workhorse
DelegateRef = make_VT("DelegateRef", pytype=internals.Delegate,
                      constructor_func=DelegateRef_ctor)


def delegate_is_a(x, typ):
    if isinstance(x, BlobPtr):
        if not internals.is_delegate(x):
            return False
        del_ref = to_delegate_imp(x)
    elif isinstance(x, DelegateRef):
        del_ref = x
    else:
        return False

    # Get to here, x is at least some kind of delegate
    if len(typ._d["absorbed"]) == 0:
        return True

    raise NotImplementedError("TODO")
    # Could probably just do a isinstance on the absorbed part, but not if
    # there's more interesting types going in there.

Delegate = make_VT("Delegate", is_a_func=delegate_is_a)

def instancecheck_Delegate(self, other):
    raise NotImplementedError("TODO")
    